import os
import json
import traceback
from http import HTTPStatus
from typing import Dict, Iterable, List

from flask import request, Response
from flask_restx import Resource

from mindsdb.interfaces.agents.agents_controller import AgentsController
from mindsdb.interfaces.storage import db
from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.utilities.log import getLogger
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError


logger = getLogger(__name__)


AGENT_QUICK_RESPONSE = "I understand your request. I'm working on a detailed response for you."


def create_agent(project_name, name, agent):
    if name is None:
        return http_error(
            HTTPStatus.BAD_REQUEST,
            'Missing field',
            'Missing "name" field for agent'
        )

    if 'model_name' not in agent:
        return http_error(
            HTTPStatus.BAD_REQUEST,
            'Missing field',
            'Missing "model_name" field for agent'
        )

    model_name = agent['model_name']
    provider = agent.get('provider')
    params = agent.get('params', {})
    skills = agent.get('skills', [])

    agents_controller = AgentsController()

    try:
        existing_agent = agents_controller.get_agent(name, project_name=project_name)
    except ValueError:
        # Project must exist.
        return http_error(
            HTTPStatus.NOT_FOUND,
            'Project not found',
            f'Project with name {project_name} does not exist'
        )
    if existing_agent is not None:
        return http_error(
            HTTPStatus.CONFLICT,
            'Agent already exists',
            f'Agent with name {name} already exists. Please choose a different one.'
        )

    try:
        created_agent = agents_controller.add_agent(
            name=name,
            project_name=project_name,
            model_name=model_name,
            skills=skills,
            provider=provider,
            params=params
        )
        return created_agent.as_dict(), HTTPStatus.CREATED
    except ValueError:
        # Model or skill doesn't exist.
        return http_error(
            HTTPStatus.NOT_FOUND,
            'Resource not found',
            f'The model "{model_name}" or skills "{skills}" do not exist. Please ensure that the names are correct and try again.'
        )
    except NotImplementedError:
        # Free users trying to create agent.
        return http_error(
            HTTPStatus.UNAUTHORIZED,
            'Unavailable to free users',
            f'The model "{model_name}" or skills "{skills}" do not exist. Please ensure that the names are correct and try again.'
        )


@ns_conf.route('/<project_name>/agents')
class AgentsResource(Resource):
    @ns_conf.doc('list_agents')
    @api_endpoint_metrics('GET', '/agents')
    def get(self, project_name):
        ''' List all agents '''
        session = SessionController()
        try:
            all_agents = session.agents_controller.get_agents(project_name)
        except EntityNotExistsError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist')
        return [a.as_dict() for a in all_agents]

    @ns_conf.doc('create_agent')
    @api_endpoint_metrics('POST', '/agents')
    def post(self, project_name):
        '''Create a agent'''

        # Check for required parameters.
        if 'agent' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "agent" parameter in POST body'
            )

        agent = request.json['agent']

        name = agent.get('name')
        return create_agent(project_name, name, agent)


@ns_conf.route('/<project_name>/agents/<agent_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('agent_name', 'Name of the agent')
class AgentResource(Resource):
    @ns_conf.doc('get_agent')
    @api_endpoint_metrics('GET', '/agents/agent')
    def get(self, project_name, agent_name):
        '''Gets an agent by name'''
        session = SessionController()
        try:
            existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Agent not found',
                    f'Agent with name {agent_name} does not exist'
                )
            return existing_agent.as_dict()
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

    @ns_conf.doc('update_agent')
    @api_endpoint_metrics('PUT', '/agents/agent')
    def put(self, project_name, agent_name):
        '''Updates an agent by name, creating one if it doesn't exist'''

        # Check for required parameters.
        if 'agent' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "agent" parameter in POST body'
            )
        agents_controller = AgentsController()

        try:
            existing_agent_record = agents_controller.get_agent(agent_name, project_name=project_name)
        except (ValueError, EntityNotExistsError):
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )
        if existing_agent_record is None:
            raise Exception

        agent = request.json['agent']
        name = agent.get('name', None)
        model_name = agent.get('model_name', None)
        skills_to_add = agent.get('skills_to_add', [])
        skills_to_remove = agent.get('skills_to_remove', [])
        skills_to_rewrite = agent.get('skills', [])
        provider = agent.get('provider')
        params = agent.get('params', None)

        # Agent must not exist with new name.
        if name is not None and name != agent_name:
            agent_with_new_name = agents_controller.get_agent(name, project_name=project_name)
            if agent_with_new_name is not None:
                return http_error(
                    HTTPStatus.CONFLICT,
                    'Agent already exists',
                    f'Agent with name {name} already exists. Please choose a different one.'
                )

        if existing_agent_record is None:
            # Create
            return create_agent(project_name, name, agent)

        # Update
        try:
            # Prepare the params dictionary
            if params is None:
                params = {}

            # Check if any of the skills to be added is of type 'retrieval'
            session = SessionController()
            skills_controller = session.skills_controller
            retrieval_skill_added = False
            if len(skills_to_add) > 0:
                skills_names = [x['name'] if isinstance(x, dict) else x for x in skills_to_add]
                retrieval_skill_added = any(
                    skills_controller.get_skill(skill_name).type == 'retrieval'
                    for skill_name in skills_names
                    if skills_controller.get_skill(skill_name) is not None
                )
            elif len(skills_to_rewrite) > 0:
                retrieval_skill_added = any(
                    skills_controller.get_skill(skill_meta['name']).type == 'retrieval'
                    for skill_meta in skills_to_rewrite
                    if skills_controller.get_skill(skill_meta['name']) is not None
                )

            if retrieval_skill_added and 'mode' not in params:
                params['mode'] = 'retrieval'

            updated_agent = agents_controller.update_agent(
                agent_name,
                project_name=project_name,
                name=name,
                model_name=model_name,
                skills_to_add=skills_to_add,
                skills_to_remove=skills_to_remove,
                skills_to_rewrite=skills_to_rewrite,
                provider=provider,
                params=params
            )

            return updated_agent.as_dict()
        except EntityExistsError as e:
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Resource should not exists',
                str(e)
            )
        except EntityNotExistsError as e:
            # Agent or skill doesn't exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Resource not found',
                str(e)
            )
        except ValueError as e:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Wrong arguments',
                str(e)
            )

    @ns_conf.doc('delete_agent')
    @api_endpoint_metrics('DELETE', '/agents/agent')
    def delete(self, project_name, agent_name):
        '''Deletes a agent by name'''
        agents_controller = AgentsController()

        try:
            existing_agent = agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Agent not found',
                    f'Agent with name {agent_name} does not exist'
                )
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        agents_controller.delete_agent(agent_name, project_name=project_name)
        return '', HTTPStatus.NO_CONTENT


def _completion_event_generator(
        agent_name: str,
        messages: List[Dict],
        project_name: str) -> Iterable[str]:
    logger.info(f"Starting completion event generator for agent {agent_name}")

    def json_serialize(data):
        return f'data: {json.dumps(data)}\n\n'

    quick_response_message = {
        'role': 'assistant',
        'content': AGENT_QUICK_RESPONSE
    }
    yield json_serialize({"quick_response": True, "messages": [quick_response_message]})
    logger.info("Quick response sent")

    try:
        # Populate API key by default if not present.
        session = SessionController()
        existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
        if not existing_agent.params:
            existing_agent.params = {}
        existing_agent.params['openai_api_key'] = existing_agent.params.get('openai_api_key',
                                                                            os.getenv('OPENAI_API_KEY'))
        # Have to commit/flush here so DB isn't locked while streaming.
        db.session.commit()

        if (
            'mode' not in existing_agent.params
            and any(rel.skill.type == 'retrieval' for rel in existing_agent.skills_relationships)
        ):
            existing_agent.params['mode'] = 'retrieval'

        completion_stream = session.agents_controller.get_completion(
            existing_agent,
            messages,
            project_name=project_name,
            tools=[],
            stream=True
        )

        for chunk in completion_stream:
            if isinstance(chunk, str) and chunk.startswith('data: '):
                # The chunk is already formatted correctly, yield it as is
                yield chunk
            elif isinstance(chunk, dict):
                if 'error' in chunk:
                    # Handle error chunks
                    logger.error(f"Error in completion stream: {chunk['error']}")
                    yield json_serialize({"error": chunk['error']})
                elif chunk.get('type') == 'context':
                    # Handle context message
                    yield json_serialize({"type": "context", "content": chunk.get('content')})
                elif chunk.get('type') == 'sql':
                    # Handle SQL query message
                    yield json_serialize({"type": "sql", "content": chunk.get('content')})
                else:
                    # Chunk should already be formatted by agent stream.
                    yield json_serialize(chunk)
            else:
                # For any other unexpected chunk types
                yield json_serialize({"output": str(chunk)})

            logger.debug(f"Streamed chunk: {str(chunk)[:100]}...")

        logger.info("Completion stream finished")

    except Exception as e:
        error_message = f"Error in completion event generator: {str(e)}"
        logger.error(error_message)
        logger.error(traceback.format_exc())
        yield json_serialize({"error": error_message})

    finally:
        yield json_serialize({"type": "end"})


@ns_conf.route('/<project_name>/agents/<agent_name>/completions/stream')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('agent_name', 'Name of the agent')
class AgentCompletionsStream(Resource):
    @ns_conf.doc('agent_completions_stream')
    @api_endpoint_metrics('POST', '/agents/agent/completions/stream')
    def post(self, project_name, agent_name):
        logger.info(f"Received streaming request for agent {agent_name} in project {project_name}")

        # Check for required parameters.
        if 'messages' not in request.json:
            logger.error("Missing 'messages' parameter in request body")
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "messages" parameter in POST body'
            )

        session = SessionController()
        try:
            existing_agent = session.agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                logger.error(f"Agent {agent_name} not found in project {project_name}")
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Agent not found',
                    f'Agent with name {agent_name} does not exist'
                )
        except ValueError as e:
            logger.error(f"Project {project_name} not found: {str(e)}")
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        messages = request.json['messages']

        try:
            gen = _completion_event_generator(
                agent_name,
                messages,
                project_name
            )
            logger.info(f"Starting streaming response for agent {agent_name}")
            return Response(gen, mimetype='text/event-stream')
        except Exception as e:
            logger.error(f"Error during streaming for agent {agent_name}: {str(e)}")
            logger.error(traceback.format_exc())
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR,
                'Streaming error',
                f'An error occurred during streaming: {str(e)}'
            )


@ns_conf.route('/<project_name>/agents/<agent_name>/completions')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('agent_name', 'Name of the agent')
class AgentCompletions(Resource):
    @ns_conf.doc('agent_completions')
    @api_endpoint_metrics('POST', '/agents/agent/completions')
    def post(self, project_name, agent_name):
        '''Queries an agent given a list of messages'''
        # Check for required parameters.
        if 'messages' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "messages" parameter in POST body'
            )
        agents_controller = AgentsController()

        try:
            existing_agent = agents_controller.get_agent(agent_name, project_name=project_name)
            if existing_agent is None:
                return http_error(
                    HTTPStatus.NOT_FOUND,
                    'Agent not found',
                    f'Agent with name {agent_name} does not exist'
                )
        except ValueError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )

        # Add OpenAI API key to agent params if not already present.
        if not existing_agent.params:
            existing_agent.params = {}
        existing_agent.params['openai_api_key'] = existing_agent.params.get('openai_api_key', os.getenv('OPENAI_API_KEY'))

        # set mode to `retrieval` if agent has a skill of type `retrieval` and mode is not set
        if (
            'mode' not in existing_agent.params
            and any(rel.skill.type == 'retrieval' for rel in existing_agent.skills_relationships)
        ):
            existing_agent.params['mode'] = 'retrieval'

        messages = request.json['messages']

        completion = agents_controller.get_completion(
            existing_agent,
            messages,
            project_name=project_name,
            # Don't need to include backoffice_db related tools into this endpoint.
            # Underlying handler (e.g. Langchain) will handle default tools like mdb_read, mdb_write, etc.
            tools=[]
        )

        output_col = agents_controller.assistant_column
        model_output = completion.iloc[-1][output_col]
        trace_id = completion.iloc[-1]['trace_id']

        response = {
            'message': {
                'content': model_output,
                'role': 'assistant',
                'trace_id': trace_id
            }
        }

        if existing_agent.params.get('return_context', False):
            context = []
            if 'context' in completion.columns:
                try:
                    last_context = completion.iloc[-1]['context']
                    if last_context:
                        context = json.loads(last_context)
                except (json.JSONDecodeError, IndexError) as e:
                    logger.error(f'Error decoding context: {e}')
                    pass  # Keeping context as an empty list in case of error

            response['message']['context'] = context

        return response



import time
import types
from concurrent.futures import ThreadPoolExecutor
import contextvars


class ContextThreadPoolExecutor(ThreadPoolExecutor):
    '''Handles copying context variables to threads created by ThreadPoolExecutor'''
    def __init__(self, max_workers=None):
        self.context = contextvars.copy_context()
        # ThreadPoolExecutor does not propagate context to threads by default, so we need a custom initializer.
        super().__init__(max_workers=max_workers, initializer=self._set_child_context)

    def _set_child_context(self):
        for var, value in self.context.items():
            var.set(value)


def execute_in_threads(func, tasks, thread_count=3, queue_size_k=1.5):
    """
    Should be used as generator.
    Can accept input tasks as generator and keep queue size the same to not overflow the RAM

    :param func: callable, function to execute in threads
    :param tasks: generator or iterable, list of input for function
    :param thread_count: number of threads
    :param queue_size_k: how a queue for workers is bigger than count of threads
    :return: yield results
    """
    executor = ContextThreadPoolExecutor(max_workers=thread_count)

    queue_size = int(thread_count * queue_size_k)

    if not isinstance(tasks, types.GeneratorType):
        tasks = iter(tasks)

    futures = None
    while futures is None or len(futures) > 0:
        if futures is None:
            futures = []

        # add new portion
        for i in range(queue_size):
            try:
                args = next(tasks)
                futures.append(executor.submit(func, *args))
            except StopIteration:
                break

        # save results
        for task in futures:
            if task.done():
                yield task.result()

        # remove completed tasks
        futures[:] = [t for t in futures if not t.done()]

        time.sleep(0.1)
    executor.shutdown(wait=False)



import overpy

from mindsdb.integrations.handlers.openstreetmap_handler.openstreetmap_tables import (OpenStreetMapNodeTable,
                                                                                      OpenStreetMapWayTable, OpenStreetMapRelationTable)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log
from mindsdb_sql_parser import parse_sql

logger = log.getLogger(__name__)


class OpenStreetMapHandler(APIHandler):
    """The OpenStreetMap handler implementation."""

    def __init__(self, name: str, **kwargs):
        """Registers all API tables and prepares the handler for an API connection.

        Args:
            name: (str): The handler name to use
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        nodes_data = OpenStreetMapNodeTable(self)
        self._register_table("nodes", nodes_data)

        ways_data = OpenStreetMapWayTable(self)
        self._register_table("ways", ways_data)

        relations_data = OpenStreetMapRelationTable(self)
        self._register_table("relations", relations_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.

        Returns:
            StatusResponse: connection object
        """
        if self.is_connected is True:
            return self.connection

        api_session = overpy.Overpass()

        self.connection = api_session

        self.is_connected = True

        return self.connection

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns:
            HandlerStatusResponse
        """
        response = StatusResponse(False)

        try:
            api_session = self.connect()
            if api_session is not None:
                response.success = True
        except Exception as e:
            logger.error('Error connecting to OpenStreetMap!')
            response.error_message = str(e)

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> Response:
        """Execute a native query on the handler.

        Args:
            query (str): The query to execute.

        Returns:
            Response: The response from the query.
        """
        ast = parse_sql(query)
        return self.query(ast)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    credentials={
        'type': ARG_TYPE.PATH,
        'description': 'The path to the credentials file. If not specified, the default credentials are used.'
    },
    merchant_id={
        'type': ARG_TYPE.STR,
        'description': 'The merchant ID for the Google Content API.'
    },
)

connection_args_example = OrderedDict(
    credentials='/path/to/credentials.json',
    merchant_id='1234567890'
)



from opentelemetry import metrics
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.metrics.export import (
    MetricExporter,
    PeriodicExportingMetricReader,
)


def setup_meter(resource: Resource, exporter: MetricExporter) -> None:
    """
    Setup OpenTelemetry metrics
    """

    metric_reader = PeriodicExportingMetricReader(exporter=exporter)
    provider = MeterProvider(resource=resource, metric_readers=[metric_reader])

    # Sets the global default meter provider
    metrics.set_meter_provider(provider)



"""
    Wrapper around MindsDB's executor and integration controller following the implementation of the original
    langchain.sql_database.SQLDatabase class to partly replicate its behavior.
"""
from typing import Any, Iterable, List, Optional

from mindsdb.utilities import log
from langchain_community.utilities import SQLDatabase
from mindsdb.interfaces.skills.sql_agent import SQLAgent

logger = log.getLogger(__name__)


class MindsDBSQL(SQLDatabase):
    @staticmethod
    def custom_init(
        sql_agent: 'SQLAgent'
    ) -> 'MindsDBSQL':
        instance = MindsDBSQL()
        instance._sql_agent = sql_agent
        return instance

    """ Can't modify signature, as LangChain does a Pydantic check."""
    def __init__(
        self,
        engine: Optional[Any] = None,
        schema: Optional[str] = None,
        metadata: Optional[Any] = None,
        ignore_tables: Optional[List[str]] = None,
        include_tables: Optional[List[str]] = None,
        sample_rows_in_table_info: int = 3,
        indexes_in_table_info: bool = False,
        custom_table_info: Optional[dict] = None,
        view_support: bool = True,
        max_string_length: int = 300,
        lazy_table_reflection: bool = False,
    ):
        pass

    @property
    def dialect(self) -> str:
        return 'mindsdb'

    @property
    def table_info(self) -> str:
        """Information about all tables in the database."""
        return self._sql_agent.get_table_info()

    def get_usable_table_names(self) -> Iterable[str]:
        return self._sql_agent.get_usable_table_names()

    def get_table_info_no_throw(self, table_names: Optional[List[str]] = None) -> str:
        for i in range(len(table_names)):
            if '$START$' in table_names[i]:
                table_names[i] = table_names[i].partition('$START$')[-1]
            if '$END$' in table_names[i]:
                table_names[i] = table_names[i].partition('$END$')[0]
            table_names[i] = table_names[i].strip(' ')
        return self._sql_agent.get_table_info_safe(table_names)

    def run_no_throw(self, command: str, fetch: str = "all") -> str:
        return self._sql_agent.query_safe(command)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the DocumentDB server.',
        'required': True,
        'label': 'User',
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the DocumentDB server.',
        'required': True,
        'label': 'Password',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the DocumentDB server.',
        'required': True,
        'label': 'Database',
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the DocumentDB server. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.',
        'required': True,
        'label': 'Host',
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the DocumentDB server. Must be an integer.',
        'required': True,
        'label': 'Port',
    },
    kwargs={
        'type': ARG_TYPE.DICT,
        'description': 'Additional parameters of DocumentDB same as MongoDB.',
        'required': False,
        'label': 'Kwargs',
    },
)


connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=27017,
    username='documentdb',
    password='password',
    database='database',
)



"""create_chat_bots_history_table

Revision ID: b5bf593ba659
Revises: 9d6271bb2c38
Create Date: 2023-06-19 09:56:37.108680

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b5bf593ba659'
down_revision = '9d6271bb2c38'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'chat_bots_history',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('chat_bot_id', sa.Integer(), nullable=False),
        sa.Column('type', sa.String()),
        sa.Column('text', sa.String()),
        sa.Column('user', sa.String()),
        sa.Column('destination', sa.String()),
        sa.Column('sent_at', sa.DateTime()),
        sa.Column('error', sa.String()),
        sa.PrimaryKeyConstraint('id'),
    )


def downgrade():
    op.drop_table('chat_bots_history')



import pandas as pd
from typing import List

from google.analytics.admin_v1beta import ListConversionEventsRequest, ConversionEvent, CreateConversionEventRequest, \
    UpdateConversionEventRequest, DeleteConversionEventRequest
from mindsdb_sql_parser import Constant
from mindsdb_sql_parser import ast
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class ConversionEventsTable(APITable):

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Gets all conversion events from google analytics property.

        Args:
            query (ast.Select): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """
        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        # Get the page size from the conditions.
        params = {}
        for op, arg1, arg2 in conditions:
            if arg1 == 'page_size':
                params[arg1] = arg2
            else:
                raise NotImplementedError

        # Get the order by from the query.
        if query.order_by is not None:
            raise NotImplementedError

        if query.limit is not None:
            raise NotImplementedError

        # Get the conversion events from the Google Analytics Admin API.
        conversion_events = pd.DataFrame(columns=self.get_columns())
        result = self.get_conversion_events(params=params)
        conversion_events_data = self.extract_conversion_events_data(result.conversion_events)
        events = self.concat_dataframes(conversion_events, conversion_events_data)

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(events) == 0:
            events = pd.DataFrame([], columns=selected_columns)
        else:
            events.columns = self.get_columns()
            for col in set(events.columns).difference(set(selected_columns)):
                events = events.drop(col, axis=1)
        return events

    def insert(self, query: ast.Insert):
        """
        Inserts a conversion event into your GA4 property.

        Args:
            query (ast.Insert): SQL query to parse.
        """
        columns = [col.name for col in query.columns]

        supported_columns = {'event_name', 'countingMethod'}
        if not set(columns).issubset(supported_columns):
            unsupported_columns = set(columns).difference(supported_columns)
            raise ValueError(
                "Unsupported columns for create conversion event: "
                + ", ".join(unsupported_columns)
            )
        params = {}

        for row in query.values:
            params = dict(zip(columns, row))

        # get params values of a type <Constant>
        if isinstance(params['countingMethod'], str):
            params['countingMethod'] = int(params['countingMethod'])
        elif isinstance(params['countingMethod'], Constant):
            params['countingMethod'] = params['countingMethod'].value
        else:
            params['countingMethod'] = params['countingMethod']

        if isinstance(params['event_name'], Constant):
            params['event_name'] = params['event_name'].value
        else:
            params['event_name'] = params['event_name']

        # Insert the conversion event into the Google Analytics Admin API.
        conversion_events = pd.DataFrame(columns=self.get_columns())
        result = self.create_conversion_event(params=params)
        conversion_events_data = self.extract_conversion_events_data([result])
        self.concat_dataframes(conversion_events, conversion_events_data)

    def update(self, query: ast.Update):
        """
        Updates a conversion event into your GA4 property.

        Args:
            query (ast.Update): SQL query to parse.
        """
        # Get the values from the query.
        values = query.update_columns.items()
        data_list = list(values)
        # Get the conversion event data from the values.
        params = {}
        for col, val in zip(query.update_columns, data_list):
            params[col] = val

        conditions = extract_comparison_conditions(query.where)
        for op, arg1, arg2 in conditions:
            if arg1 == 'name':
                if op == '=':
                    params['name'] = arg2
                else:
                    raise NotImplementedError
            else:
                raise NotImplementedError

        # get params values of a type <Constant>
        params['countingMethod'] = params['countingMethod'][1].value

        # Update the conversion event in the Google Analytics Admin API.
        conversion_events = pd.DataFrame(columns=self.get_columns())
        result = self.update_conversion_event(params=params)
        conversion_events_data = self.extract_conversion_events_data([result])
        self.concat_dataframes(conversion_events, conversion_events_data)

    def delete(self, query: ast.Delete):
        """
        Deletes a conversion event into your GA4 property.

        Args:
            query (ast.Delete): SQL query to parse.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        for op, arg1, arg2 in conditions:
            if op == 'or':
                raise NotImplementedError('OR is not supported')
            if arg1 == 'name':
                if op == '=':
                    self.delete_conversion_event(params={'name': arg2})
                else:
                    raise NotImplementedError(f'Unknown op: {op}')
            else:
                raise NotImplementedError(f'Unknown clause: {arg1}')

    def get_conversion_events(self, params: dict = None):
        """
        List all conversion events in your GA4 property
        Args:
            params (dict): query parameters
        Returns:
            ConversionEvent objects
        """
        service = self.handler.connect()
        page_token = None
        url = self.handler.get_api_url('properties')

        while True:
            request = ListConversionEventsRequest(parent=url,
                                                  page_token=page_token, **params)
            result = service.list_conversion_events(request)

            page_token = result.next_page_token
            if not page_token:
                break
        return result

    def create_conversion_event(self, params: dict = None):
        """
        Create a conversion event in your property.
        Args:
            params (dict): query parameters
        Returns:
            ConversionEvent object
        """
        service = self.handler.connect()
        url = self.handler.get_api_url('properties')

        conversion_event = ConversionEvent(
            event_name=params['event_name'],
            counting_method=params['countingMethod']
        )
        request = CreateConversionEventRequest(conversion_event=conversion_event,
                                               parent=url)
        result = service.create_conversion_event(request)

        return result

    def update_conversion_event(self, params: dict = None):
        """
        Update a conversion event in your property.
        Args:
            params (dict): query parameters
        Returns:
            ConversionEvent object
        """
        service = self.handler.connect()

        conversion_event = ConversionEvent(
            name=params['name'],
            counting_method=params['countingMethod']
        )
        request = UpdateConversionEventRequest(conversion_event=conversion_event,
                                               update_mask='*')
        result = service.update_conversion_event(request)

        return result

    def delete_conversion_event(self, params: dict = None):
        """
        Delete a conversion event in your property.
        Args:
            params (dict): query parameters
        """
        service = self.handler.connect()
        request = DeleteConversionEventRequest(name=params['name'])
        service.delete_conversion_event(request)

    @staticmethod
    def extract_conversion_events_data(conversion_events):
        """
        Extract conversion events data and return a list of lists.
        Args:
            conversion_events: List of ConversionEvent objects
        Returns:
            List of lists containing conversion event data
        """
        conversion_events_data = []
        for conversion_event in conversion_events:
            data_row = [
                conversion_event.name,
                conversion_event.event_name,
                conversion_event.create_time,
                conversion_event.deletable,
                conversion_event.custom,
                conversion_event.ConversionCountingMethod(conversion_event.counting_method).name,
            ]
            conversion_events_data.append(data_row)
        return conversion_events_data

    def concat_dataframes(self, existing_df, data):
        """
        Concatenate existing DataFrame with new data.
        Args:
            existing_df: Existing DataFrame
            data: New data to be added to the DataFrame
        Returns:
            Concatenated DataFrame
        """
        return pd.concat(
            [existing_df, pd.DataFrame(data, columns=self.get_columns())],
            ignore_index=True
        )

    def get_columns(self) -> List[str]:
        """
        Gets all columns to be returned in pandas DataFrame responses

        Returns:
        List[str]: List of columns
        """
        return [
            'name',
            'event_name',
            'create_time',
            'deletable',
            'custom',
            'countingMethod',
        ]



from typing import List, Dict, Optional, Any
import pandas as pd
from langchain_text_splitters import RecursiveCharacterTextSplitter
import hashlib
import asyncio


from mindsdb.integrations.utilities.rag.splitters.file_splitter import (
    FileSplitter,
    FileSplitterConfig,
)

from mindsdb.interfaces.agents.langchain_agent import create_chat_model

from mindsdb.interfaces.knowledge_base.preprocessing.models import (
    PreprocessingConfig,
    ProcessedChunk,
    PreprocessorType,
    ContextualConfig,
    Document,
    TextChunkingConfig,
)
from mindsdb.utilities import log

from langchain_core.documents import Document as LangchainDocument

logger = log.getLogger(__name__)


class DocumentPreprocessor:
    """Base class for document preprocessing"""

    RESERVED_METADATA_FIELDS = {
        "content",
        "id",
        "embeddings",
        "original_doc_id",
        "chunk_index",
    }

    def __init__(self):
        """Initialize preprocessor"""
        self.splitter = None  # Will be set by child classes

    def process_documents(self, documents: List[Document]) -> List[ProcessedChunk]:
        """Base implementation - should be overridden by child classes"""
        raise NotImplementedError("Subclasses must implement process_documents")

    def _split_document(self, doc: Document) -> List[Document]:
        """Split document into chunks while preserving metadata"""
        if self.splitter is None:
            raise ValueError("Splitter not configured")

        # Convert to langchain Document for splitting
        langchain_doc = LangchainDocument(
            page_content=doc.content, metadata=doc.metadata or {}
        )
        # Split and convert back to our Document type
        split_docs = self.splitter.split_documents([langchain_doc])
        return [
            Document(content=split_doc.page_content, metadata=split_doc.metadata)
            for split_doc in split_docs
        ]

    def _get_source(self) -> str:
        """Get the source identifier for this preprocessor"""
        return self.__class__.__name__

    def to_dataframe(self, chunks: List[ProcessedChunk]) -> pd.DataFrame:
        """Convert processed chunks to dataframe format"""
        return pd.DataFrame([chunk.model_dump() for chunk in chunks])

    @staticmethod
    def dict_to_document(data: Dict[str, Any]) -> Document:
        """Convert a dictionary to a Document object"""
        return Document(
            id=data.get("id"),
            content=data.get("content", ""),
            embeddings=data.get("embeddings"),
            metadata=data.get("metadata", {}),
        )

    def _generate_deterministic_id(
        self, content: str, content_column: str = None, provided_id: str = None
    ) -> str:
        """Generate a deterministic ID based on content and column"""
        if provided_id is not None:
            return f"{provided_id}_{content_column}"

        id_string = f"content={content}_column={content_column}"
        return hashlib.sha256(id_string.encode()).hexdigest()

    def _generate_chunk_id(
        self,
        content: str,
        chunk_index: Optional[int] = None,
        content_column: str = None,
        provided_id: str = None,
    ) -> str:
        """Generate deterministic ID for a chunk"""
        base_id = self._generate_deterministic_id(content, content_column, provided_id)
        chunk_id = (
            f"{base_id}_chunk_{chunk_index}" if chunk_index is not None else base_id
        )
        logger.debug(f"Generated chunk ID: {chunk_id} for content hash: {base_id}")
        return chunk_id

    def _prepare_chunk_metadata(
        self,
        doc_id: Optional[str],
        chunk_index: Optional[int],
        base_metadata: Optional[Dict] = None,
    ) -> Dict:
        """Centralized method for preparing chunk metadata"""
        metadata = base_metadata or {}

        # Always preserve original document ID
        if doc_id is not None:
            metadata["original_doc_id"] = doc_id

        # Add chunk index only for multi-chunk cases
        if chunk_index is not None:
            metadata["chunk_index"] = chunk_index

        # Always set source
        metadata["source"] = self._get_source()

        return metadata


class ContextualPreprocessor(DocumentPreprocessor):
    """Contextual preprocessing implementation that enhances document chunks with context"""

    DEFAULT_CONTEXT_TEMPLATE = """
<document>
{{WHOLE_DOCUMENT}}
</document>
Here is the chunk we want to situate within the whole document
<chunk>
{{CHUNK_CONTENT}}
</chunk>
Please give a short succinct context to situate this chunk within the overall document for the purposes of improving search retrieval of the chunk. Answer only with the succinct context and nothing else."""

    def __init__(self, config: ContextualConfig):
        """Initialize with contextual configuration"""
        super().__init__()
        self.config = config
        self.splitter = FileSplitter(
            FileSplitterConfig(
                chunk_size=config.chunk_size, chunk_overlap=config.chunk_overlap
            )
        )
        self.llm = create_chat_model(
            {
                "model_name": self.config.llm_config.model_name,
                "provider": self.config.llm_config.provider,
                **self.config.llm_config.params,
            }
        )
        self.context_template = config.context_template or self.DEFAULT_CONTEXT_TEMPLATE
        self.summarize = self.config.summarize

    def _prepare_prompts(
        self, chunk_contents: list[str], full_documents: list[str]
    ) -> list[str]:
        prompts = [
            self.context_template.replace("{{WHOLE_DOCUMENT}}", full_document)
            for full_document in full_documents
        ]
        prompts = [
            prompt.replace("{{CHUNK_CONTENT}}", chunk_content)
            for prompt, chunk_content in zip(prompts, chunk_contents)
        ]

        return prompts

    def _generate_context(
        self, chunk_contents: list[str], full_documents: list[str]
    ) -> list[str]:
        """Generate contextual description for a chunk using LLM"""
        prompts = self._prepare_prompts(chunk_contents, full_documents)

        # Check if LLM supports async
        if hasattr(self.llm, 'abatch'):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                response = loop.run_until_complete(self._async_generate(prompts))
            finally:
                loop.close()
        else:
            # Use sync batch for non-async LLMs
            response = [resp.content for resp in self.llm.batch(prompts)]
        return response

    async def _async_generate(self, prompts: list[str]) -> list[str]:
        """Helper for async LLM generation"""
        return [resp.content for resp in await self.llm.abatch(prompts)]

    def _split_document(self, doc: Document) -> List[Document]:
        """Split document into chunks while preserving metadata"""
        # Use base class implementation
        return super()._split_document(doc)

    def process_documents(self, documents: List[Document]) -> List[ProcessedChunk]:
        chunks_list = []
        doc_index_list = []
        chunk_index_list = []
        processed_chunks = []

        for doc_index, doc in enumerate(documents):
            # Get content_column from metadata if available
            content_column = (
                doc.metadata.get("content_column") if doc.metadata else None
            )

            # Ensure document has an ID
            if doc.id is None:
                doc.id = self._generate_deterministic_id(doc.content, content_column)

            # Skip empty or whitespace-only content
            if not doc.content or not doc.content.strip():
                continue

            chunk_docs = self._split_document(doc)

            # Single chunk case
            if len(chunk_docs) == 1:
                chunk_doc = chunk_docs[0]
                if not chunk_doc.content or not chunk_doc.content.strip():
                    continue
                else:
                    chunks_list.append(chunk_doc)
                    doc_index_list.append(doc_index)
                    chunk_index_list.append(0)

            else:
                # Multiple chunks case
                for i, chunk_doc in enumerate(chunk_docs):
                    if not chunk_doc.content or not chunk_doc.content.strip():
                        continue
                    else:
                        chunks_list.append(chunk_doc)
                        doc_index_list.append(doc_index)
                        chunk_index_list.append(i)

        # Generate contexts
        doc_contents = [documents[i].content for i in doc_index_list]
        chunk_contents = [chunk_doc.content for chunk_doc in chunks_list]
        contexts = self._generate_context(chunk_contents, doc_contents)

        for context, chunk_doc, chunk_index, doc_index in zip(
            contexts, chunks_list, chunk_index_list, doc_index_list
        ):
            processed_content = (
                context if self.summarize else f"{context}\n\n{chunk_doc.content}"
            )
            doc = documents[doc_index]

            # Initialize metadata
            metadata = {}
            if doc.metadata:
                metadata.update(doc.metadata)

            # Pass through doc.id and content_column
            content_column = (
                doc.metadata.get("content_column") if doc.metadata else None
            )
            chunk_id = self._generate_chunk_id(
                processed_content,
                chunk_index,
                content_column=content_column,
                provided_id=doc.id,
            )
            processed_chunks.append(
                ProcessedChunk(
                    id=chunk_id,
                    content=processed_content,  # Use the content with context
                    embeddings=doc.embeddings,
                    metadata=self._prepare_chunk_metadata(doc.id, None, metadata),
                )
            )

        return processed_chunks


class TextChunkingPreprocessor(DocumentPreprocessor):
    """Default text chunking preprocessor using RecursiveCharacterTextSplitter"""

    def __init__(self, config: Optional[TextChunkingConfig] = None):
        """Initialize with text chunking configuration"""
        super().__init__()
        self.config = config or TextChunkingConfig()
        self.splitter = RecursiveCharacterTextSplitter(
            chunk_size=self.config.chunk_size,
            chunk_overlap=self.config.chunk_overlap,
            length_function=self.config.length_function,
            separators=self.config.separators,
        )

    def _split_document(self, doc: Document) -> List[Document]:
        """Split document into chunks while preserving metadata"""
        # Use base class implementation
        return super()._split_document(doc)

    def process_documents(self, documents: List[Document]) -> List[ProcessedChunk]:
        processed_chunks = []

        for doc in documents:
            # Get content_column from metadata if available
            content_column = (
                doc.metadata.get("content_column") if doc.metadata else None
            )

            # Ensure document has an ID
            if doc.id is None:
                doc.id = self._generate_deterministic_id(doc.content, content_column)

            # Skip empty or whitespace-only content
            if not doc.content or not doc.content.strip():
                continue

            chunk_docs = self._split_document(doc)

            # Single chunk case
            if len(chunk_docs) == 1:
                chunk_doc = chunk_docs[0]
                if not chunk_doc.content or not chunk_doc.content.strip():
                    continue

                # Initialize metadata
                metadata = {}
                if doc.metadata:
                    metadata.update(doc.metadata)

                # Pass through doc.id and content_column
                id = self._generate_chunk_id(
                    chunk_doc.content, content_column=content_column, provided_id=doc.id
                )
                processed_chunks.append(
                    ProcessedChunk(
                        id=id,
                        content=chunk_doc.content,
                        embeddings=doc.embeddings,
                        metadata=self._prepare_chunk_metadata(doc.id, None, metadata),
                    )
                )
            else:
                # Multiple chunks case
                for i, chunk_doc in enumerate(chunk_docs):
                    if not chunk_doc.content or not chunk_doc.content.strip():
                        continue

                    # Initialize metadata
                    metadata = {}
                    if doc.metadata:
                        metadata.update(doc.metadata)

                    # Pass through doc.id and content_column
                    chunk_id = self._generate_chunk_id(
                        chunk_doc.content,
                        i,
                        content_column=content_column,
                        provided_id=doc.id,
                    )
                    processed_chunks.append(
                        ProcessedChunk(
                            id=chunk_id,
                            content=chunk_doc.content,
                            embeddings=doc.embeddings,
                            metadata=self._prepare_chunk_metadata(doc.id, i, metadata),
                        )
                    )

        return processed_chunks


class PreprocessorFactory:
    """Factory for creating preprocessors based on configuration"""

    @staticmethod
    def create_preprocessor(
        config: Optional[PreprocessingConfig] = None,
    ) -> DocumentPreprocessor:
        """
        Create appropriate preprocessor based on configuration
        : param config: Preprocessing configuration
        : return: Configured preprocessor instance
        : raises ValueError: If unknown preprocessor type specified
        """
        if config is None:
            return TextChunkingPreprocessor()

        if config.type == PreprocessorType.CONTEXTUAL:
            return ContextualPreprocessor(
                config.contextual_config or ContextualConfig()
            )
        elif config.type == PreprocessorType.TEXT_CHUNKING:
            return TextChunkingPreprocessor(config.text_chunking_config)

        raise ValueError(f"Unknown preprocessor type: {config.type}")



from collections import OrderedDict
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

connection_args = OrderedDict(
    cos_hmac_access_key_id={
        "type": ARG_TYPE.PWD,
        "description": "IBM COS HMAC Access Key ID.",
        "required": True,
        "label": "HMAC Access Key ID",
        "secret": True,
    },
    cos_hmac_secret_access_key={
        "type": ARG_TYPE.PWD,
        "description": "IBM COS HMAC Secret Access Key.",
        "required": True,
        "label": "HMAC Secret Access Key",
        "secret": True,
    },
    cos_endpoint_url={
        "type": ARG_TYPE.STR,
        "description": "IBM COS Endpoint URL (e.g., https://s3.eu-gb.cloud-object-storage.appdomain.cloud).",
        "required": True,
        "label": "Endpoint URL",
    },
    bucket={
        "type": ARG_TYPE.STR,
        "description": "IBM COS Bucket Name (Optional).",
        "required": False,
        "label": "Bucket Name",
    },
)

connection_args_example = OrderedDict(
    cos_hmac_access_key_id="YOUR_HMAC_ACCESS_KEY_ID",
    cos_hmac_secret_access_key="YOUR_HMAC_SECRET_ACCESS_KEY",
    cos_endpoint_url="https://s3.eu-gb.cloud-object-storage.appdomain.cloud",
    bucket="YOUR_BUCKET_NAME",
)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    dsn={
        "type": ARG_TYPE.STR,
        "description": "The data source name (DSN) for the Oracle database.",
        "required": False,
        "label": "Data Source Name (DSN)",
    },
    host={
        "type": ARG_TYPE.STR,
        "description": "The hostname, IP address, or URL of the Oracle server.",
        "required": False,
        "label": "Host",
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "The port number for connecting to the Oracle database. Default is 1521.",
        "required": False,
        "label": "Port",
    },
    sid={
        "type": ARG_TYPE.STR,
        "description": "The system identifier (SID) of the Oracle database.",
        "required": False,
        "label": "SID",
    },
    service_name={
        "type": ARG_TYPE.STR,
        "description": "The service name of the Oracle database.",
        "required": False,
        "label": "Service Name",
    },
    user={
        "type": ARG_TYPE.STR,
        "description": "The username for the Oracle database.",
        "required": True,
        "label": "User",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password for the Oracle database.",
        "secret": True,
        "required": True,
        "label": "Password",
    },
    disable_oob={
        "type": ARG_TYPE.BOOL,
        "description": "The boolean parameter to disable out-of-band breaks. Default is `false`.",
        "required": False,
        "label": "Disable OOB",
    },
    auth_mode={
        "type": ARG_TYPE.STR,
        "description": "The authorization mode to use.",
        "required": False,
        "label": "Auth Mode",
    },
)

connection_args_example = OrderedDict(
    host="127.0.0.1",
    port=1521,
    user="admin",
    password="password",
    sid="ORCL",
)



import unittest

from mindsdb_sql_parser import parse_sql

from mindsdb.integrations.handlers.notion_handler.notion_handler import NotionHandler
from mindsdb.integrations.handlers.notion_handler.notion_table import (
    NotionDatabaseTable,
    NotionBlocksTable,
    NotionCommentsTable,
    NotionPagesTable,
)


class NotionHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "api_token": "secret_KHTlOzUN5fIwVlb1euOpBa4lwcJA7jEALoBGDRrlATx"
            }
        }
        cls.handler = NotionHandler("test_notion_handler", **cls.kwargs)
        cls.db_table = NotionDatabaseTable(cls.handler)
        cls.pages_table = NotionPagesTable(cls.handler)
        cls.blocks_table = NotionBlocksTable(cls.handler)
        cls.comments_table = NotionCommentsTable(cls.handler)

    def test_check_connection(self):
        status = self.handler.check_connection()
        self.assertTrue(status.success)

    def test_select_database(self):
        query = "SELECT * FROM database WHERE database_id = '21510b8a953c4d62958c9907f3cf9f87'"
        ast = parse_sql(query)
        res = self.db_table.select(ast)
        self.assertFalse(res.empty)

    def test_select_page(self):
        query = "SELECT * FROM pages WHERE page_id = '70f28e55416b4dfe8588aa175ecae63a'"
        ast = parse_sql(query)
        res = self.pages_table.select(ast)
        self.assertFalse(res.empty)

    def test_select_blocks(self):
        query = (
            "SELECT * FROM blocks WHERE block_id = '6d1480e0bf4b46e1a71be093f105d654'"
        )
        ast = parse_sql(query)
        res = self.blocks_table.select(ast)
        self.assertFalse(res.empty)

    def test_select_comment(self):
        query = (
            "SELECT * FROM comments where block_id = '169fa742a8374fe9a516caecfb33432a'"
        )
        ast = parse_sql(query)
        res = self.comments_table.select(ast)
        self.assertFalse(res.empty)


if __name__ == "__main__":
    unittest.main()



import os
import requests
import time
from typing import Dict, Optional

import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.libs.llm.utils import get_completed_prompts
from mindsdb.utilities import log
from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.utilities.config import Config

logger = log.getLogger(__name__)

LEONARDO_API_BASE = 'https://cloud.leonardo.ai/api/rest/v1'


class LeonardoAIHandler(BaseMLEngine):
    """
    This integration seamlessly combines MindsDB and Leonardo AI to create a powerful
    AI-driven solution for creative content generation.

    Content Generation with Leonardo AI: Harness the power of advanced generative
    models for creative content production. From realistic images to artistic text, Leonardo
    AI opens up new possibilities for content creators.
    """
    name = "leonardo_ai"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.all_models = []
        self.default_model = '6bef9f1b-29cb-40c7-b9df-32b51c1f67d3'
        self.base_api = LEONARDO_API_BASE
        self.rate_limit = 50
        self.max_batch_size = 5  # default value

    def create(self, target: str, args=None, **kwargs):

        args['target'] = target

        if "using" not in args:
            raise Exception(
                "Leornardo Engine requires a USING clause!"
            )

        self.model_storage.json_set("args", args)
        api_key = self._get_leonardo_api_key(args, self.engine_storage)  # fetch api key

        # check if API key is valid
        self.connection = requests.get(
            "https://cloud.leonardo.ai/api/rest/v1/me",
            headers={
                "accept": "application/json",
                "authorization": f"Bearer {api_key}"
            }
        )

        # if valid, check if the model is valid
        try:
            if (self.connection.status_code == 200):
                # get all the available models
                available_models = self._get_platform_model(args)
                if not args['using']['model']:
                    args['using']['model'] = self.default_model
                elif args['using']['model'] not in available_models:    # if invalid model_id is provided
                    raise Exception("Invalid Model ID. Please use a valid Model")
            else:
                raise Exception("Unable to make connection, please verify the API key.")
        except Exception:
            raise Exception("Auth Connection Error, please the modelId or API key")

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None, **kwargs) -> pd.DataFrame:

        pred_args = args['predict_params'] if args else {}
        args = self.model_storage.json_get("args")

        prompt_template = pred_args.get('prompt_template', args.get('prompt_template', 'Generate a picture of {{{{text}}}}'))

        # prepare prompts
        prompts, empty_prompt_id = get_completed_prompts(prompt_template, df)
        df['__mdb_prompt'] = prompts

        # generate picture based on the given prompt
        result_df = pd.DataFrame(self.predict_answer(prompts))
        return result_df

    def _get_leonardo_api_key(self, args, engine_storage: HandlerStorage, strict=True):
        """
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. LEONARDO_API_KEY env variable
            4. leonardo.api_key setting in config.json
        """

        # 1
        if "api_key" in args["using"]:
            return args["using"]["api_key"]
        # 2
        connection_args = engine_storage.get_connection_args()
        if "api_key" in connection_args:
            return connection_args["api_key"]
        # 3
        api_key = os.getenv("LEONARDO_API_KEY")
        if api_key is not None:
            return api_key
        # 4
        config = Config()
        leonardo_cfg = config.get("leonardo", {})
        if "api_key" in leonardo_cfg:
            return leonardo_cfg["api_key"]

        if strict:
            raise Exception(
                'Missing API key "api_key". Either re-create this ML_ENGINE specifying the `api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.'
            )

    def _get_platform_model(self, args):
        """
        Returns a list of available model based on the API key provided
        """
        model_ids = []

        args = self.model_storage.json_get('args')
        api_key = self._get_leonardo_api_key(args, self.engine_storage)

        self.connection = requests.get(
            "https://cloud.leonardo.ai/api/rest/v1/platformModels",
            headers={
                "accept": "application/json",
                "authorization": f"Bearer {api_key}"
            }
        )

        models = self.connection.json()

        # extract the model ids from the response
        model_ids = [model['id'] for model in models['custom_models']]

        return model_ids

    def predict_answer(self, prompts, **kwargs):
        """
        Generates pictures based on the prompts and returns URLs with few variations.

        Request Flow:
            - POST request with a prompt is sent
            - `generation_id` is created for the request
            - POST request will take couple of seconds to generate the picture, till then the process will be kept busy with a simple math calculation.
            - New GET request with the `generation_id` will fetch the generated pictures as URLs
        """
        args = self.model_storage.json_get('args')
        tmp_args = args['using']
        height = tmp_args.get('height', 512)
        width = tmp_args.get('width', 512)
        api_key = self._get_leonardo_api_key(args, self.engine_storage)  # fetch API key
        generation_id = ''

        # Endpoint URL
        generation_url = "https://cloud.leonardo.ai/api/rest/v1/generations"

        post_headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "authorization": f"Bearer {api_key}"
        }

        get_headers = {
            "accept": "application/json",
            "authorization": f"Bearer {api_key}"
        }

        # payload
        generation_payload = {
            "height": height,
            "modelId": args['using']['model'],
            "prompt": f"{prompts}",
            "width": width,
        }

        # Make a POST request to generate the image
        response_generation = requests.post(generation_url, json=generation_payload, headers=post_headers)
        generation_data = response_generation.json()

        # Wait for 15 seconds

        time.sleep(15)

        # extract generationID from the response
        generation_id = generation_data['sdGenerationJob']['generationId']

        # ENDPOINT GET URL
        retrieve_url = f"https://cloud.leonardo.ai/api/rest/v1/generations/{generation_id}"

        # GET request to retrieve image URLs
        response_retrieve = requests.get(retrieve_url, headers=get_headers)
        retrieve_data = response_retrieve.json()

        # extract URLs from the response
        generated_images = retrieve_data["generations_by_pk"]["generated_images"]
        image_urls = [image["url"] for image in generated_images]

        url_dicts = []

        for url in image_urls:
            url_dicts.append({'url': url})

        img_urls = pd.DataFrame(url_dicts, columns=['url'])
        return img_urls

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        args = self.model_storage.json_get('args')
        model, target = args['using']['model'], args['target']
        prompt_template = args.get('prompt_template', 'Generate a picture of {{{{text}}}}')

        if attribute == "features":
            return pd.DataFrame([[target, prompt_template]], columns=['target_column', 'mindsdb_prompt_template'])
        elif attribute == "metadata":
            api_key = self._get_leonardo_api_key(args, self.engine_storage)
            return pd.DataFrame([[target, api_key, model]], columns=['target', 'api_key', 'model_name'])
        else:
            tables = ['args', 'api_key']
            return pd.DataFrame(tables, columns=['tables'])



from typing import Dict, Optional

import dill
import pandas as pd
import polars as pl

from mindsdb.integrations.libs.base import BaseMLEngine


class PopularityRecommenderHandler(BaseMLEngine):
    """
    Integration with polar based popularity recommender.
    """

    name = "popularity-recommender"

    def create(
        self,
        target: str,
        df: pd.DataFrame = None,
        meta_data: Dict = None,
        args: Optional[Dict] = None,
    ) -> None:

        args = args["using"]

        interaction_data = pl.from_pandas(df)

        args["ave_per_item_user"] = (
            interaction_data.get_column(args["user_id"])
            .value_counts()
            .mean()["count"][0]
        )

        popularity = (
            interaction_data.get_column(args["item_id"])
            .value_counts()
            .sort("count", descending=True)
            .get_column(args["item_id"])
            .head(
                int(args["n_recommendations"] * args["ave_per_item_user"])
            )  # to ensure there are enough to predict
            .to_pandas()
            .reset_index()
            .to_dict(orient="list")
        )

        self.model_storage.file_set("interaction", dill.dumps(df))
        self.model_storage.json_set("popularity", popularity)
        self.model_storage.json_set("args", args)

    def predict(self, df=None, args: Optional[dict] = None):

        args = self.model_storage.json_get("args")
        popularity = self.model_storage.json_get("popularity")
        interaction = dill.loads(self.model_storage.file_get("interaction"))

        global_popularity = [*popularity.values()][1]

        if df is not None:
            # get recommendations for specific users if specified
            user_ids = df[args["user_id"]].unique().tolist()

            interaction_data = pl.from_pandas(interaction).filter(
                pl.col(args["user_id"]).is_in(user_ids)
            )

        else:
            # get recommendations for all users
            interaction_data = pl.from_pandas(interaction)

        # aggregate over user
        interacted_items = (
            interaction_data.groupby(args["user_id"])
            .agg(pl.col(args["item_id"]).alias("items"))
            .with_columns(
                pl.lit([global_popularity]).alias("popular_items"),
            )
        ).lazy()

        df = (
            (
                interacted_items.join(
                    (
                        interacted_items.explode("popular_items")
                        .filter(pl.col("popular_items").is_in("items").is_not())
                        .groupby(args["user_id"])
                        .agg(recommended="popular_items")
                    ),
                    on=args["user_id"],
                ).select(
                    [
                        pl.col(args["user_id"]),
                        pl.col("recommended").list.head(args["n_recommendations"]),
                    ]
                )
            )
            .explode("recommended")
            .collect()
            .to_pandas()
        )

        return df



import numpy as np
import pandas as pd
from typing import Text, Tuple, Dict, List, Optional, Any

from mindsdb.utilities import log

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.libs.llm.utils import get_completed_prompts
from mindsdb.integrations.libs.api_handler_exceptions import MissingConnectionParams
from mindsdb.integrations.handlers.bedrock_handler.utilities import create_amazon_bedrock_client
from mindsdb.integrations.handlers.bedrock_handler.settings import AmazonBedrockHandlerEngineConfig, AmazonBedrockHandlerModelConfig


logger = log.getLogger(__name__)


class AmazonBedrockHandler(BaseMLEngine):
    """
    This handler handles connection and inference with the Amazon Bedrock API.
    """

    name = 'bedrock'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True

    def create_engine(self, connection_args: Dict) -> None:
        """
        Validates the AWS credentials provided on creation of an engine.

        Args:
            connection_args (Dict): The parameters of the engine.

        Raises:
            Exception: If the handler is not configured with valid API credentials.
        """
        connection_args = {k.lower(): v for k, v in connection_args.items()}
        AmazonBedrockHandlerEngineConfig(**connection_args)

    def create(self, target, args: Dict = None, **kwargs: Any) -> None:
        """
        Creates a model by validating the model configuration and saving it to the storage.

        Args:
            target (Text): The target column name.
            args (Dict): The parameters of the model.
            kwargs (Any): Other keyword arguments.

        Raises:
            Exception: If the model is not configured with valid parameters.

        Returns:
            None
        """
        if 'using' not in args:
            raise MissingConnectionParams("Amazon Bedrock engine requires a USING clause! Refer to its documentation for more details.")
        else:
            model_args = args['using']
            # Replace 'model_id' with 'id' to match the Amazon Bedrock handler model configuration.
            # This is done to avoid the Pydantic warning regarding conflicts with the protected 'model_' namespace.
            if 'model_id' in model_args:
                model_args['id'] = model_args['model_id']
                del model_args['model_id']

            handler_model_config = AmazonBedrockHandlerModelConfig(**model_args, connection_args=self.engine_storage.get_connection_args())

            # Save the model configuration to the storage.
            handler_model_params = handler_model_config.model_dump()
            logger.info(f"Saving model configuration to storage: {handler_model_params}")

            args['target'] = target
            args['handler_model_params'] = handler_model_params
            self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> pd.DataFrame:
        """
        Makes predictions using a model by invoking the Amazon Bedrock API.

        Args:
            df (pd.DataFrame): The input data to invoke the model with.
            args (Dict): The parameters passed when making predictions.

        Raises:
            ValueError: If the input data does not match the configuration of the model.

        Returns:
            pd.DataFrame: The input data with the predicted values in a new column.
        """
        args = self.model_storage.json_get('args')
        handler_model_params = args['handler_model_params']
        mode = handler_model_params['mode']
        model_id = handler_model_params['id']
        inference_config = handler_model_params.get('inference_config')
        target = args['target']

        if mode == 'default':
            prompts, empty_prompt_ids = self._prepare_data_for_default_mode(df, args)
            predictions = self._predict_for_default_mode(model_id, prompts, inference_config)

            # Fill the empty predictions with None.
            for i in sorted(empty_prompt_ids):
                predictions.insert(i, None)

        elif mode == 'conversational':
            prompt, total_questions = self._prepare_data_for_conversational_mode(df, args)
            prediction = self._predict_for_conversational_mode(model_id, prompt, inference_config)

            # Create a list of None values for the total number of questions and replace the last one with the prediction.
            predictions = [None] * total_questions
            predictions[-1] = prediction

        pred_df = pd.DataFrame(predictions, columns=[target])
        return pred_df

    def _prepare_data_for_default_mode(self, df: pd.DataFrame, args: Dict) -> List[Dict]:
        """
        Prepares the input data for the default mode of the Amazon Bedrock handler.
        A separate prompt is prepared for each question.

        Args:
            df (pd.DataFrame): The input data to invoke the model with.
            args (Dict): The parameters of the model.

        Returns:
            List[Dict]: The prepared prompts for invoking the Amazon Bedrock API. The model will be invoked for each prompt.
        """
        handler_model_params = args['handler_model_params']
        question_column = handler_model_params.get('question_column')
        context_column = handler_model_params.get('context_column')
        prompt_template = handler_model_params.get('prompt_template')

        if question_column is not None:
            questions, empty_prompt_ids = self._prepare_data_with_question_and_context_columns(
                df,
                question_column,
                context_column
            )

        elif prompt_template is not None:
            questions, empty_prompt_ids = self._prepare_data_with_prompt_template(df, prompt_template)

        # Prepare the prompts.
        questions = [question for i, question in enumerate(questions) if i not in empty_prompt_ids]
        prompts = [{"role": "user", "content": [{"text": question}]} for question in questions]

        return prompts, empty_prompt_ids

    def _prepare_data_for_conversational_mode(self, df: pd.DataFrame, args: Dict) -> Tuple[List[Dict], int]:
        """
        Prepares the input data for the conversational mode of the Amazon Bedrock handler.
        A single prompt is prepared for all the questions.

        Args:
            df (pd.DataFrame): The input data to invoke the model with.
            args (Dict): The parameters of the model.

        Returns:
            Tuple[List[Dict], int]: The prepared prompt for invoking the Amazon Bedrock API and the total number of questions.
            The model will be invoked once using this prompt which contains all the questions.
            The total number of questions is used to produce the final list of predictions.
        """
        handler_model_params = args['handler_model_params']
        question_column = handler_model_params.get('question_column')
        context_column = handler_model_params.get('context_column')
        prompt_template = handler_model_params.get('prompt_template')

        if question_column is not None:
            questions, empty_prompt_ids = self._prepare_data_with_question_and_context_columns(
                df,
                question_column,
                context_column
            )

        if prompt_template is not None:
            questions, empty_prompt_ids = self._prepare_data_with_prompt_template(df, prompt_template)

        # Prepare the prompts.
        questions = [question for i, question in enumerate(questions) if i not in empty_prompt_ids]
        prompt = [{"role": "user", "content": [{"text": question} for question in questions]}]

        # Get the total number of questions; including the empty ones.
        total_questions = len(df)

        return prompt, total_questions

    def _prepare_data_with_question_and_context_columns(self, df: pd.DataFrame, question_column: Text, context_column: Text = None) -> Tuple[List[Text], List[int]]:
        """
        Prepares the input data with question and context columns.

        Args:
            df (pd.DataFrame): The input data to invoke the model with.
            question_column (Text): The column containing the questions.
            context_column (Text): The column containing the context.

        Returns:
            Tuple[List[Text], List[int]]: The questions to build the prompts for invoking the Amazon Bedrock API and the empty prompt IDs.
        """
        if question_column not in df.columns:
            raise ValueError(f"Column {question_column} not found in the dataframe!")

        if context_column and context_column not in df.columns:
            raise ValueError(f"Column {context_column} not found in the dataframe!")

        if context_column:
            empty_prompt_ids = np.where(
                df[[context_column, question_column]]
                .isna()
                .all(axis=1)
                .values
            )[0]
            contexts = list(df[context_column].apply(lambda x: str(x)))
            questions_without_context = list(df[question_column].apply(lambda x: str(x)))

            questions = [
                f'Context: {c}\nQuestion: {q}\nAnswer: '
                for c, q in zip(contexts, questions_without_context)
            ]

        else:
            questions = list(df[question_column].apply(lambda x: str(x)))
            empty_prompt_ids = np.where(
                df[[question_column]].isna().all(axis=1).values
            )[0]

        return questions, empty_prompt_ids

    def _prepare_data_with_prompt_template(self, df: pd.DataFrame, prompt_template: Text) -> Tuple[List[Text], List[int]]:
        """
        Prepares the input data with a prompt template.

        Args:
            df (pd.DataFrame): The input data to invoke the model with.
            prompt_template (Text): The base prompt template to use.

        Returns:
            Tuple[List[Text], List[int]]: The questions to build the prompts for invoking the Amazon Bedrock API and the empty prompt IDs.
        """
        questions, empty_prompt_ids = get_completed_prompts(prompt_template, df)

        return questions, empty_prompt_ids

    def _predict_for_default_mode(self, model_id: Text, prompts: List[Text], inference_config: Dict) -> List[Text]:
        """
        Makes predictions for the default mode of the Amazon Bedrock handler using the prepared prompts.

        Args:
            model_id (Text): The ID of the model in Amazon Bedrock.
            prompts (List[Text]): The prepared prompts for invoking the Amazon Bedrock API.
            inference_config (Dict): The inference configuration supported by the Amazon Bedrock API.

        Returns:
            List[Text]: The predictions made by the Amazon Bedrock API.
        """
        predictions = []
        bedrock_runtime_client = create_amazon_bedrock_client(
            'bedrock-runtime',
            **self.engine_storage.get_connection_args()
        )

        for prompt in prompts:
            response = bedrock_runtime_client.converse(
                modelId=model_id,
                messages=[prompt],
                inferenceConfig=inference_config
            )
            predictions.append(
                response["output"]["message"]["content"][0]["text"]
            )

        return predictions

    def _predict_for_conversational_mode(self, model_id: Text, prompt: List[Text], inference_config: Dict) -> Text:
        """
        Makes a prediction for the conversational mode of the Amazon Bedrock handler using the prepared prompt.

        Args:
            model_id (Text): The ID of the model in Amazon Bedrock.
            prompts (List[Text]): Prepared prompts for invoking the Amazon Bedrock API.
            inference_config (Dict): Inference configuration supported by the Amazon Bedrock API.

        Returns:
            Text: The prediction made by the Amazon Bedrock API.
        """
        bedrock_runtime_client = create_amazon_bedrock_client(
            'bedrock-runtime',
            **self.engine_storage.get_connection_args()
        )

        response = bedrock_runtime_client.converse(
            modelId=model_id,
            messages=prompt,
            inferenceConfig=inference_config
        )

        return response["output"]["message"]["content"][0]["text"]

    def describe(self, attribute: Optional[Text] = None) -> pd.DataFrame:
        """
        Get the metadata or arguments of a model.

        Args:
            attribute (Optional[Text]): Attribute to describe. Can be 'args' or 'metadata'.

        Returns:
            pd.DataFrame: Model metadata or model arguments.
        """
        args = self.model_storage.json_get('args')

        if attribute == 'args':
            del args['handler_model_params']
            return pd.DataFrame(args.items(), columns=['key', 'value'])

        elif attribute == 'metadata':
            model_id = args['handler_model_params']['id']
            try:
                bedrock_client = create_amazon_bedrock_client(
                    'bedrock',
                    **self.engine_storage.get_connection_args()
                )
                meta = bedrock_client.get_foundation_model(modelIdentifier=model_id)['modelDetails']
            except Exception as e:
                meta = {'error': str(e)}
            return pd.DataFrame(dict(meta).items(), columns=['key', 'value'])

        else:
            tables = ['args', 'metadata']
            return pd.DataFrame(tables, columns=['tables'])



from mindsdb_sql_parser import ast
from typing import Text, List, Optional

from .exceptions import UnsupportedColumnException

from mindsdb.integrations.utilities.handlers.query_utilities.base_query_utilities import BaseQueryParser
from mindsdb.integrations.utilities.handlers.query_utilities.base_query_utilities import BaseQueryExecutor


class UPDATEQueryParser(BaseQueryParser):
    """
    Parses an UPDATE query into its component parts.

    Parameters
    ----------
    query : ast.Update
        Given SQL UPDATE query.
    supported_columns : List[Text], Optional
        List of columns supported by the table for updating.
    """
    def __init__(self, query: ast.Update, supported_columns: Optional[List[Text]] = None):
        super().__init__(query)
        self.supported_columns = supported_columns

    def parse_query(self):
        """
        Parses a SQL UPDATE statement into its components: the columns and values to update as a dictionary, and the WHERE conditions.
        """
        values_to_update = self.parse_set_clause()
        where_conditions = self.parse_where_clause()

        return values_to_update, where_conditions

    def parse_set_clause(self):
        """
        Parses the SET clause of the query and returns a dictionary of columns and values to update.
        """
        values = list(self.query.update_columns.items())

        values_to_update = {}
        for value in values:
            if self.supported_columns:
                if value[0] not in self.supported_columns:
                    raise UnsupportedColumnException(f"Unsupported column: {value[0]}")

            values_to_update[value[0]] = value[1].value

        return values_to_update


class UPDATEQueryExecutor(BaseQueryExecutor):
    """
    Executes an UPDATE query.

    Parameters
    ----------
    df : pd.DataFrame
        Given table.
    where_conditions : List[List[Text]]
        WHERE conditions of the query.

    NOTE: This class DOES NOT update the relevant records of the entity for you, it will simply return the records that need to be updated based on the WHERE conditions.

          This class expects all of the entities to be passed in as a DataFrane and filters out the relevant records based on the WHERE conditions.
          Because all of the records need to be extracted to be passed in as a DataFrame, this class is not very computationally efficient.
          Therefore, DO NOT use this class if the API/SDK that you are using supports updating records in bulk.
    """



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    weaviate_url={
        "type": ARG_TYPE.STR,
        "description": "weaviate url/ local endpoint",
        "required": False,
    },
    weaviate_api_key={
        "type": ARG_TYPE.STR,
        "description": "weaviate API KEY",
        "required": False,
        "secret": True
    },
    persistence_directory={
        "type": ARG_TYPE.STR,
        "description": "persistence directory for weaviate",
        "required": False,
    },
)

connection_args_example = OrderedDict(
    weaviate_url="http://localhost:8080",
    weaviate_api_key="<api_key>",
    persistence_directory="db_path",
)



from typing import Optional

import pandas as pd
from impala import dbapi as db , sqlalchemy as SA

from mindsdb_sql_parser import parse_sql
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


logger = log.getLogger(__name__)


class ImpalaHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Impala statements.
    """

    name = 'impala'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        super().__init__(name)

        
        self.parser = parse_sql
        self.dialect = 'impala'
        self.kwargs = kwargs
        self.connection_data = connection_data

        self.connection = None
        self.is_connected = False



    def connect(self):
        if self.is_connected is True:
            return self.connection

        config = {
            'host': self.connection_data.get('host'),
            'port': self.connection_data.get('port',21050),
            'user': self.connection_data.get('user'),
            'password': self.connection_data.get('password'),
            'database': self.connection_data.get('database'),
            
            
        }

        connection = db.connect(**config)
        self.is_connected = True
        self.connection = connection
        return self.connection

    def disconnect(self):
        if self.is_connected is False:
            return
        self.connection.close()
        self.is_connected = False
        return

    def check_connection(self) -> StatusResponse:
       
        result = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            result.success = connection is not None
        except Exception as e:
            logger.error(f'x x x Error connecting to Impala {self.connection_data["database"]}, {e}!')
            result.error_message = str(e)

        if result.success is True and need_to_close:
            self.disconnect()
        if result.success is False and self.is_connected is True:
            self.is_connected = False

        return result

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in Impala
        :return: returns the records from the current recordset
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)
                result = cur.fetchall()
                if cur.has_result_set:
                    
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        pd.DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                connection.commit()
            except Exception as e:
                logger.error(f'Error running query: {query} on {self.connection_data["database"]}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )
                # connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        renderer = SqlalchemyRender(SA.ImpalaDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in Impala
        """
        q = "SHOW TABLES;"
        result= self.native_query(q)
        df = result.data_frame.rename(columns={'name': 'TABLE_NAME'})
        result.data_frame = df

        return result

    def get_columns(self, table_name: str) -> Response:
        """
        Show details about the table
        """
        q = f"DESCRIBE {table_name};"

        result= self.native_query(q)
        df = result.data_frame.iloc[:,0:2].rename(columns={'name': 'COLUMN_NAME', 'type': 'Data_Type'})
        result.data_frame = df

        return result



class dtype:
    """
    Definitions of all data types currently supported. Dtypes currently supported include:


    - **Numerical**: Data that should be represented in the form of a number. Currently ``integer``, ``float``, and ``quantity`` are supported.
    - **Categorical**: Data that represents a class or label and is discrete. Currently ``binary``, ``categorical``, and ``tags`` are supported.
    - **Date/Time**: Time-series data that is temporal/sequential. Currently ``date``, and ``datetime`` are supported.
    - **Text**: Data that can be considered as language information.  Currently ``short_text``, and ``rich_text`` are supported. Short text has a small vocabulary (~ 100 words) and is generally a limited number of characters. Rich text is anything with greater complexity.
    - **Complex**: Data types that require custom techniques. Currently ``audio``, ``video`` and ``image`` are available, but highly experimental.
    - **Array**: Data in the form of a sequence where order must be preserved. ``tsarray`` dtypes are for "normal" columns that will be transformed to arrays at a row-level because they will be treated as time series.
    - **Miscellaneous**: Miscellaneous data descriptors include ``empty``, an explicitly unknown value versus ``invalid``, a data type not currently supported.
    
    Custom data types may be implemented here as a flag for subsequent treatment and processing. You are welcome to include your own definitions, so long as they do not override the existing type names (alternatively, if you do, please edit subsequent parts of the preprocessing pipeline to correctly indicate how you want to deal with these data types).
    """ # noqa

    # Numerical type data
    integer = "integer"
    float = "float"
    quantity = "quantity"

    # Categorical type data
    binary = "binary"
    categorical = "categorical"
    tags = "tags"

    # Dates and Times (time-series)
    date = "date"
    datetime = "datetime"

    # Text
    short_text = "short_text"
    rich_text = "rich_text"

    # Complex Data types
    image = "image"
    audio = "audio"
    video = "video"

    # Series/Sequences
    num_array = "num_array"
    cat_array = "cat_array"
    num_tsarray = 'num_tsarray'
    cat_tsarray = 'cat_tsarray'

    # Misc (Unk/NaNs)
    empty = "empty"
    invalid = "invalid"



from mindsdb_sql_parser import parse_sql
from mindsdb.api.executor.planner import utils as planner_utils

import mindsdb.utilities.profiler as profiler
from mindsdb.api.executor.sql_query.result_set import Column
from mindsdb.api.executor.sql_query import SQLQuery
from mindsdb.api.executor.command_executor import ExecuteCommands
from mindsdb.api.mysql.mysql_proxy.utilities import ErSqlSyntaxError
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class Executor:
    """This class stores initial and intermediate params
    between different steps of query execution. And it is also
    creates a separate instance of ExecuteCommands to execute the current
    query step.

    IMPORTANT: A public API of this class is a contract.
    And there are at least 2 classes strongly depend on it:
        ExecuctorClient
        ExecutorService.
    These classes do the same work as Executor when
    MindsDB works in 'modularity' mode.
    Thus please make sure that IF you change the API,
    you must update the API of these two classes as well!"""

    def __init__(self, session, sqlserver):
        self.session = session
        self.sqlserver = sqlserver

        self.query = None

        # returned values
        # all this attributes needs to be added in
        # self.json() method
        self.columns = []
        self.params = []
        self.data = None
        self.state_track = None
        self.server_status = None
        self.is_executed = False
        self.error_message = None
        self.error_code = None

        self.sql = ""
        self.sql_lower = ""

        context = {'connection_id': self.sqlserver.connection_id}
        self.command_executor = ExecuteCommands(self.session, context)

    def change_default_db(self, new_db):
        self.command_executor.change_default_db(new_db)

    def stmt_prepare(self, sql):

        self.parse(sql)

        # if not params
        params = planner_utils.get_query_params(self.query)
        if len(params) == 0:
            # execute immediately
            self.do_execute()

        else:
            # plan query
            # TODO less complex.
            #  planner is inside SQLQuery now.

            sqlquery = SQLQuery(self.query, session=self.session, execute=False)

            sqlquery.prepare_query()

            self.params = [
                Column(
                    alias=p.value,
                    type="str",
                    name=p.value,
                )
                for p in params
            ]

            # TODO:
            #   select * from mindsdb.models doesn't invoke prepare_steps and columns_list is empty
            self.columns = sqlquery.columns_list

    def stmt_execute(self, param_values):
        if self.is_executed:
            return

        # fill params
        self.query = planner_utils.fill_query_params(self.query, param_values)

        # execute query
        self.do_execute()

    @profiler.profile()
    def query_execute(self, sql):
        self.parse(sql)
        self.do_execute()

    @profiler.profile()
    def parse(self, sql):
        self.sql = sql
        sql_lower = sql.lower()
        self.sql_lower = sql_lower.replace("`", "")

        try:
            self.query = parse_sql(sql)
        except Exception as mdb_error:
            # not all statements are parsed by parse_sql
            logger.warning('Failed to parse SQL query')
            logger.debug(f'Query that cannot be parsed: {sql}')

            raise ErSqlSyntaxError(
                f"The SQL statement cannot be parsed - {sql}: {mdb_error}"
            ) from mdb_error

            # == a place for workarounds ==
            # or run sql in integration without parsing

    @profiler.profile()
    def do_execute(self):
        # it can be already run at prepare state
        if self.is_executed:
            return

        ret = self.command_executor.execute_command(self.query)
        self.error_code = ret.error_code
        self.error_message = ret.error_message

        self.is_executed = True

        if ret.data is not None:
            self.data = ret.data
            self.columns = ret.data.columns

        self.state_track = ret.state_track



# Prepare sentry.io for error and exception tracking
import sentry_sdk
import os
from mindsdb.utilities import log

logger = log.getLogger(__name__)

# Provide your sentry.io DSN here
SENTRY_IO_DSN = os.environ.get("SENTRY_IO_DSN", "")
# Define the environment
SENTRY_IO_ENVIRONMENT = os.environ.get("SENTRY_IO_ENVIRONMENT", "local").lower()
# This is set to our SHA when deployed so we know what version this occurred in
SENTRY_IO_RELEASE = os.environ.get("SENTRY_IO_RELEASE", "local").lower()
# How often to capture traces, 1.0 means 100%.
SENTRY_IO_TRACE_SAMPLE_RATE = float(os.environ.get("SENTRY_IO_TRACE_SAMPLE_RATE", "1.0"))
# How often to capture profiling, 1.0 means 100%.
SENTRY_IO_PROFILING_SAMPLE_RATE = float(os.environ.get("SENTRY_IO_PROFILING_SAMPLE_RATE", "1.0"))
# By default we have sentry.io enabled on all envs, except for local which is disabled by default
#   If you want to enable sentry.io on local for some reason (eg: profiling) please set SENTRY_IO_FORCE_RUN to true
SENTRY_IO_DISABLED = True if (os.environ.get("SENTRY_IO_DISABLED", "false").lower() == "true" or SENTRY_IO_ENVIRONMENT == "local") else False
SENTRY_IO_FORCE_RUN = True if os.environ.get("SENTRY_IO_FORCE_RUN", "false").lower() == "true" else False


# If we're not disabled, or if we have forced sentry to run
if SENTRY_IO_DSN and (not SENTRY_IO_DISABLED or SENTRY_IO_FORCE_RUN):
    logger.info("Sentry.io enabled")
    logger.info(f"SENTRY_IO_DSN: {SENTRY_IO_DSN}")
    logger.info(f"SENTRY_IO_ENVIRONMENT: {SENTRY_IO_ENVIRONMENT}")
    logger.info(f"SENTRY_IO_RELEASE: {SENTRY_IO_RELEASE}")
    logger.info(f"SENTRY_IO_TRACE_SAMPLE_RATE: {SENTRY_IO_TRACE_SAMPLE_RATE * 100}%")
    logger.info(f"SENTRY_IO_PROFILING_SAMPLE_RATE: {SENTRY_IO_PROFILING_SAMPLE_RATE * 100}%")

    sentry_sdk.init(
        dsn=SENTRY_IO_DSN,
        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for tracing.
        traces_sample_rate=SENTRY_IO_TRACE_SAMPLE_RATE,
        # Set profiles_sample_rate to 1.0 to profile 100%
        # of sampled transactions.
        # We recommend adjusting this value in production.
        profiles_sample_rate=SENTRY_IO_PROFILING_SAMPLE_RATE,
        # What environment we're on, by default development
        environment=SENTRY_IO_ENVIRONMENT,
        # What release/image/etc we're using, injected in Helm/Kubernetes to be the image tag
        release=SENTRY_IO_RELEASE,
    )



import os
import pytest
import pandas as pd
from unittest.mock import patch

from tests.unit.ml_handlers.base_ml_test import BaseMLAPITest


@pytest.mark.skipif(os.environ.get('MDB_TEST_ANYSCALE_ENDPOINTS_API_KEY') is None, reason='Missing API key!')
class TestAnyscaleEndpoints(BaseMLAPITest):
    """
    Integration tests for Anyscale Endpoints AI engine.
    """

    def setup_method(self):
        """
        Setup test environment by creating a project and a Anyscale Endpoints engine.
        """

        super().setup_method()
        self.run_sql("CREATE DATABASE proj")
        self.run_sql(
            f"""
            CREATE ML_ENGINE anyscale_endpoints_engine
            FROM anyscale_endpoints
            USING
            anyscale_endpoints_api_key = '{self.get_api_key('MDB_TEST_ANYSCALE_ENDPOINTS_API_KEY')}';
            """
        )

    def test_create_model_raises_exception_with_invalid_model_parameter(self):
        """
        Test for invalid parameter during model creation.
        """

        self.run_sql(
            """
            CREATE MODEL proj.test_anyscale_invalid_parameter_model
            PREDICT answer
            USING
                engine='anyscale_endpoints_engine',
                model_name='this-model-does-not-exist',
                prompt_template='dummy_prompt_template';
            """
        )
        with pytest.raises(Exception):
            self.wait_predictor("proj", "test_anyscale_invalid_parameter_model")

    # TODO: Uncomment this once the handler is updated to handle unknown arguments.
    # def test_create_model_raises_exception_with_unknown_model_argument(self):
    #     """
    #     Test for unknown argument during model creation.
    #     """

    #     self.run_sql(
    #         """
    #         CREATE MODEL proj.test_anyscale_unknown_argument_model
    #         PREDICT answer
    #         USING
    #             engine='anyscale_endpoints_engine',
    #             prompt_template='dummy_prompt_template',
    #             evidently_wrong_argument='wrong value';
    #         """
    #     )
    #     with pytest.raises(Exception):
    #         self.wait_predictor("proj", "test_anyscale_unknown_argument_model")

    def test_create_model_raises_exception_with_invalid_operation_mode(self):
        """
        Test for invalid operation mode during model creation.
        """

        self.run_sql(
            """
            CREATE MODEL proj.test_anyscale_invalid_operation_mode
            PREDICT answer
            USING
                engine='anyscale_endpoints_engine',
                prompt_template='dummy_prompt_template',
                mode='invalid_mode';
            """
        )
        with pytest.raises(Exception):
            self.wait_predictor("proj", "test_anyscale_invalid_operation_mode")

    def test_select_runs_no_errors_on_completion_sentiment_analysis_single(self):
        """
        Test for a valid response to a sentiment analysis task (completion).
        """

        self.run_sql(
            """
            CREATE MODEL proj.test_anyscale_single_sa
            PREDICT sentiment
            USING
                engine='anyscale_endpoints_engine',
                model_name = 'mistralai/Mistral-7B-Instruct-v0.1',
                prompt_template = 'Classify the sentiment of the following text as one of `positive`, `neutral` or `negative`: {{text}}';
            """
        )
        self.wait_predictor("proj", "test_anyscale_single_sa")

        result_df = self.run_sql(
            """
            SELECT sentiment
            FROM proj.test_anyscale_single_sa
            WHERE text = 'I love machine learning!';
            """
        )

        assert "positive" in result_df["sentiment"].iloc[0].lower()

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_select_runs_no_errors_on_completion_sentiment_analysis_bulk(self, mock_postgres_handler):
        """
        Test for valid reponses to bulk questions in a sentiment analysis task (completion).
        """

        df = pd.DataFrame.from_dict({"text": [
            "I love machine learning!",
            "I hate slow internet connections!"
        ]})
        self.set_handler(mock_postgres_handler, name="pg", tables={"df": df})

        self.run_sql(
            """
            CREATE MODEL proj.test_anyscale_bulk_sa
            PREDICT sentiment
            USING
                engine='anyscale_endpoints_engine',
                model_name = 'mistralai/Mistral-7B-Instruct-v0.1',
                prompt_template = 'Classify the sentiment of the following text as one of `positive`, `neutral` or `negative`: {{text}}';
            """
        )
        self.wait_predictor("proj", "test_anyscale_bulk_sa")

        result_df = self.run_sql(
            """
            SELECT p.sentiment
            FROM pg.df as t
            JOIN proj.test_anyscale_bulk_sa as p;
        """
        )

        assert "positive" in result_df["sentiment"].iloc[0].lower()
        assert "negative" in result_df["sentiment"].iloc[1].lower()

    def test_select_runs_no_errors_on_chat_completion_question_answering_single(self):
        """
        Test for a valid response to a question answering task (chat completion).
        """

        self.run_sql(
            """
            CREATE MODEL proj.test_anyscale_single_qa
            PREDICT answer
            USING
                engine='anyscale_endpoints_engine',
                model_name = 'mistralai/Mistral-7B-Instruct-v0.1',
                question_column='question';
            """
        )
        self.wait_predictor("proj", "test_anyscale_single_qa")

        result_df = self.run_sql(
            """
            SELECT answer
            FROM proj.test_anyscale_single_qa
            WHERE question = 'What is the capital of Sweden?';
            """
        )

        assert "stockholm" in result_df["answer"].iloc[0].lower()

    @patch("mindsdb.integrations.handlers.postgres_handler.Handler")
    def test_select_runs_no_errors_on_chat_completion_question_answering_bulk(self, mock_postgres_handler):
        """
        Test for valid reponses to bulk questions in a question answering task (chat completion).
        """

        df = pd.DataFrame.from_dict({"question": [
            "What is the capital of Sweden?",
            "What is the second planet in the solar system?"
        ]})
        self.set_handler(mock_postgres_handler, name="pg", tables={"df": df})

        self.run_sql(
            """
            CREATE MODEL proj.test_anyscale_bulk_qa
            PREDICT answer
            USING
                engine='anyscale_endpoints_engine',
                model_name = 'mistralai/Mistral-7B-Instruct-v0.1',
                question_column='question';
            """
        )
        self.wait_predictor("proj", "test_anyscale_bulk_qa")

        result_df = self.run_sql(
            """
            SELECT p.answer
            FROM pg.df as t
            JOIN proj.test_anyscale_bulk_qa as p;
        """
        )

        assert "stockholm" in result_df["answer"].iloc[0].lower()
        assert "venus" in result_df["answer"].iloc[1].lower()


if __name__ == '__main__':
    pytest.main([__file__])



import os
import sys
import time
import atexit
import signal
import psutil
import asyncio
import secrets
import traceback
import threading
from enum import Enum
from packaging import version
from dataclasses import dataclass, field
from typing import Callable, Optional, Tuple, List

from sqlalchemy.orm.attributes import flag_modified

from mindsdb.utilities import log

logger = log.getLogger("mindsdb")
logger.debug("Starting MindsDB...")

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.utilities.config import config
from mindsdb.utilities.starters import (
    start_http, start_mysql, start_mongo, start_postgres, start_ml_task_queue, start_scheduler, start_tasks
)
from mindsdb.utilities.ps import is_pid_listen_port, get_child_pids
from mindsdb.utilities.functions import get_versions_where_predictors_become_obsolete
from mindsdb.interfaces.database.integrations import integration_controller
import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.utilities.install import install_dependencies
from mindsdb.utilities.fs import clean_process_marks, clean_unlinked_process_marks
from mindsdb.utilities.telemetry import telemetry_file_exists, disable_telemetry
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.auth import register_oauth_client, get_aws_meta_data
from mindsdb.utilities.sentry import sentry_sdk  # noqa: F401

try:
    import torch.multiprocessing as mp
except Exception:
    import multiprocessing as mp
try:
    mp.set_start_method('spawn')
except RuntimeError:
    logger.info('Torch multiprocessing context already set, ignoring...')


_stop_event = threading.Event()


class TrunkProcessEnum(Enum):
    HTTP = 'http'
    MYSQL = 'mysql'
    MONGODB = 'mongodb'
    POSTGRES = 'postgres'
    JOBS = 'jobs'
    TASKS = 'tasks'
    ML_TASK_QUEUE = 'ml_task_queue'

    @classmethod
    def _missing_(cls, value):
        print(f'"{value}" is not a valid name of subprocess')
        sys.exit(1)


@dataclass
class TrunkProcessData:
    name: str
    entrypoint: Callable
    need_to_run: bool = False
    port: Optional[int] = None
    process: Optional[mp.Process] = None
    started: bool = False
    args: Optional[Tuple] = None
    restart_on_failure: bool = False
    max_restart_count: int = 3
    max_restart_interval_seconds: int = 60

    _restart_count: int = 0
    _restarts_time: List[int] = field(default_factory=list)

    def request_restart_attempt(self) -> bool:
        """Check if the process may be restarted.
        If `max_restart_count` == 0, then there are not restrictions on restarts count or interval.
        If `max_restart_interval_seconds` == 0, then there are no time limit for restarts count.

        Returns:
            bool: `True` if the number of restarts in the interval does not exceed
        """
        if self.max_restart_count == 0:
            return True
        current_time_seconds = int(time.time())
        self._restarts_time.append(current_time_seconds)
        if self.max_restart_interval_seconds > 0:
            self._restarts_time = [
                x for x in self._restarts_time
                if x >= (current_time_seconds - self.max_restart_interval_seconds)
            ]
        if len(self._restarts_time) > self.max_restart_count:
            return False
        return True

    @property
    def should_restart(self) -> bool:
        """In case of OOM we want to restart the process. OS kill the process with code 9 on linux when an OOM occurs.
        On other OS process will be restarted regardless the code.

        Returns:
            bool: `True` if the process need to be restarted on failure
        """
        if config.is_cloud:
            return False
        if sys.platform in ('linux', 'darwin'):
            return self.restart_on_failure and self.process.exitcode == -signal.SIGKILL.value
        else:
            if self.max_restart_count == 0:
                # to prevent infinity restarts, max_restart_count should be > 0
                logger.warning('In the current OS, it is not possible to use `max_restart_count=0`')
                return False
            return self.restart_on_failure


def close_api_gracefully(trunc_processes_struct):
    _stop_event.set()
    try:
        for trunc_processes_data in trunc_processes_struct.values():
            process = trunc_processes_data.process
            if process is None:
                continue
            try:
                childs = get_child_pids(process.pid)
                for p in childs:
                    try:
                        os.kill(p, signal.SIGTERM)
                    except Exception:
                        p.kill()
                sys.stdout.flush()
                process.terminate()
                process.join()
                sys.stdout.flush()
            except psutil.NoSuchProcess:
                pass
    except KeyboardInterrupt:
        sys.exit(0)


def set_error_model_status_by_pids(unexisting_pids: List[int]):
    """Models have id of its traiing process in the 'training_metadata' field.
    If the pid does not exist, we should set the model status to "error".
    Note: only for local usage.

    Args:
        unexisting_pids (List[int]): list of 'pids' that do not exist.
    """
    predictor_records = (
        db.session.query(db.Predictor)
        .filter(
            db.Predictor.deleted_at.is_(None),
            db.Predictor.status.not_in([
                db.PREDICTOR_STATUS.COMPLETE,
                db.PREDICTOR_STATUS.ERROR
            ])
        )
        .all()
    )
    for predictor_record in predictor_records:
        predictor_process_id = (predictor_record.training_metadata or {}).get('process_id')
        if predictor_process_id in unexisting_pids:
            predictor_record.status = db.PREDICTOR_STATUS.ERROR
            if isinstance(predictor_record.data, dict) is False:
                predictor_record.data = {}
            if 'error' not in predictor_record.data:
                predictor_record.data['error'] = 'The training process was terminated for unknown reasons'
                flag_modified(predictor_record, 'data')
            db.session.commit()


def set_error_model_status_for_unfinished():
    """Set error status to any model if status not in 'complete' or 'error'
    Note: only for local usage.
    """
    predictor_records = (
        db.session.query(db.Predictor)
        .filter(
            db.Predictor.deleted_at.is_(None),
            db.Predictor.status.not_in([
                db.PREDICTOR_STATUS.COMPLETE,
                db.PREDICTOR_STATUS.ERROR
            ])
        )
        .all()
    )
    for predictor_record in predictor_records:
        predictor_record.status = db.PREDICTOR_STATUS.ERROR
        if isinstance(predictor_record.data, dict) is False:
            predictor_record.data = {}
        if 'error' not in predictor_record.data:
            predictor_record.data['error'] = 'Unknown error'
            flag_modified(predictor_record, 'data')
        db.session.commit()


def do_clean_process_marks():
    """delete unexisting 'process marks'
    """
    while _stop_event.wait(timeout=5) is False:
        unexisting_pids = clean_unlinked_process_marks()
        if not config.is_cloud and len(unexisting_pids) > 0:
            set_error_model_status_by_pids(unexisting_pids)


if __name__ == '__main__':
    # warn if less than 1Gb of free RAM
    if psutil.virtual_memory().available < (1 << 30):
        logger.warning(
            'The system is running low on memory. '
            + 'This may impact the stability and performance of the program.'
        )

    ctx.set_default()

    # ---- CHECK SYSTEM ----
    if not (sys.version_info[0] >= 3 and sys.version_info[1] >= 9):
        print("""
     MindsDB requires Python >= 3.9 to run

     Once you have supported Python version installed you can start mindsdb as follows:

     1. create and activate venv:
     python3 -m venv venv
     source venv/bin/activate

     2. install MindsDB:
     pip3 install mindsdb

     3. Run MindsDB
     python3 -m mindsdb

     More instructions in https://docs.mindsdb.com
         """)
        exit(1)

    if config.cmd_args.version:
        print(f'MindsDB {mindsdb_version}')
        sys.exit(0)

    config.raise_warnings(logger=logger)
    os.environ["MINDSDB_RUNTIME"] = "1"

    if telemetry_file_exists(config.paths['root']):
        os.environ['CHECK_FOR_UPDATES'] = '0'
        logger.info('\n x telemetry disabled! \n')
    elif os.getenv('CHECK_FOR_UPDATES', '1').lower() in ['0', 'false', 'False'] or config.is_cloud:
        disable_telemetry(config.paths['root'])
        logger.info('\n x telemetry disabled! \n')
    else:
        logger.info("✓ telemetry enabled")

    if os.environ.get("FLASK_SECRET_KEY") is None:
        os.environ["FLASK_SECRET_KEY"] = secrets.token_hex(32)

    if os.environ.get('ARROW_DEFAULT_MEMORY_POOL') is None:
        try:
            """It seems like snowflake handler have memory issue that related to pyarrow. Memory usage keep growing with
            requests. This is related to 'memory pool' that is 'mimalloc' by default: it is fastest but use a lot of ram
            """
            import pyarrow as pa
            try:
                pa.jemalloc_memory_pool()
                os.environ['ARROW_DEFAULT_MEMORY_POOL'] = 'jemalloc'
            except NotImplementedError:
                pa.system_memory_pool()
                os.environ['ARROW_DEFAULT_MEMORY_POOL'] = 'system'
        except Exception:
            pass

    db.init()
    mp.freeze_support()

    environment = config["environment"]
    if environment == "aws_marketplace":
        try:
            register_oauth_client()
        except Exception as e:
            logger.error(f"Something went wrong during client register: {e}")
    elif environment != "local":
        try:
            aws_meta_data = get_aws_meta_data()
            config.update({
                'aws_meta_data': aws_meta_data
            })
        except Exception:
            pass

    is_cloud = config.is_cloud

    if not is_cloud:
        logger.debug("Applying database migrations")
        try:
            from mindsdb.migrations import migrate
            migrate.migrate_to_head()
        except Exception as e:
            logger.error(f"Error! Something went wrong during DB migrations: {e}")

    apis = os.getenv('MINDSDB_APIS') or config.cmd_args.api

    if apis is None:  # If "--api" option is not specified, start the default APIs
        api_arr = [TrunkProcessEnum.HTTP, TrunkProcessEnum.MYSQL]
    elif apis == "":  # If "--api=" (blank) is specified, don't start any APIs
        api_arr = []
    else:  # The user has provided a list of APIs to start
        api_arr = [TrunkProcessEnum(name) for name in apis.split(',')]

    if config.cmd_args.install_handlers is not None:
        handlers_list = [s.strip() for s in config.cmd_args.install_handlers.split(",")]
        # import_meta = handler_meta.get('import', {})
        for handler_name, handler_meta in integration_controller.get_handlers_import_status().items():
            if handler_name not in handlers_list:
                continue
            import_meta = handler_meta.get("import", {})
            if import_meta.get("success") is True:
                logger.info(f"{'{0: <18}'.format(handler_name)} - already installed")
                continue
            result = install_dependencies(import_meta.get("dependencies", []))
            if result.get("success") is True:
                logger.info(
                    f"{'{0: <18}'.format(handler_name)} - successfully installed"
                )
            else:
                logger.info(
                    f"{'{0: <18}'.format(handler_name)} - error during dependencies installation: {result.get('error_message', 'unknown error')}"
                )
        sys.exit(0)

    logger.info(f"Version: {mindsdb_version}")
    logger.info(f"Configuration file: {config.config_path or 'absent'}")
    logger.info(f"Storage path: {config.paths['root']}")
    logger.debug(f"User config: {config.user_config}")
    logger.debug(f"System config: {config.auto_config}")
    logger.debug(f"Env config: {config.env_config}")

    unexisting_pids = clean_unlinked_process_marks()
    if not is_cloud:
        if len(unexisting_pids) > 0:
            set_error_model_status_by_pids(unexisting_pids)
        set_error_model_status_for_unfinished()

        integration_controller.create_permanent_integrations()

        # region Mark old predictors as outdated
        is_modified = False
        predictor_records = (
            db.session.query(db.Predictor)
            .filter(db.Predictor.deleted_at.is_(None))
            .all()
        )
        if len(predictor_records) > 0:
            (
                sucess,
                compatible_versions,
            ) = get_versions_where_predictors_become_obsolete()
            if sucess is True:
                compatible_versions = [version.parse(x) for x in compatible_versions]
                mindsdb_version_parsed = version.parse(mindsdb_version)
                compatible_versions = [x for x in compatible_versions if x <= mindsdb_version_parsed]
                if len(compatible_versions) > 0:
                    last_compatible_version = compatible_versions[-1]
                    for predictor_record in predictor_records:
                        if (
                            isinstance(predictor_record.mindsdb_version, str)
                            and version.parse(predictor_record.mindsdb_version) < last_compatible_version
                        ):
                            predictor_record.update_status = "available"
                            is_modified = True
        if is_modified is True:
            db.session.commit()
        # endregion

    clean_process_marks()

    http_api_config = config['api']['http']
    mysql_api_config = config['api']['mysql']
    trunc_processes_struct = {
        TrunkProcessEnum.HTTP: TrunkProcessData(
            name=TrunkProcessEnum.HTTP.value,
            entrypoint=start_http,
            port=http_api_config['port'],
            args=(config.cmd_args.verbose, config.cmd_args.no_studio),
            restart_on_failure=http_api_config.get('restart_on_failure', False),
            max_restart_count=http_api_config.get('max_restart_count', TrunkProcessData.max_restart_count),
            max_restart_interval_seconds=http_api_config.get(
                'max_restart_interval_seconds', TrunkProcessData.max_restart_interval_seconds
            )
        ),
        TrunkProcessEnum.MYSQL: TrunkProcessData(
            name=TrunkProcessEnum.MYSQL.value,
            entrypoint=start_mysql,
            port=mysql_api_config['port'],
            args=(config.cmd_args.verbose,),
            restart_on_failure=mysql_api_config.get('restart_on_failure', False),
            max_restart_count=mysql_api_config.get('max_restart_count', TrunkProcessData.max_restart_count),
            max_restart_interval_seconds=mysql_api_config.get(
                'max_restart_interval_seconds', TrunkProcessData.max_restart_interval_seconds
            )
        ),
        TrunkProcessEnum.MONGODB: TrunkProcessData(
            name=TrunkProcessEnum.MONGODB.value,
            entrypoint=start_mongo,
            port=config['api']['mongodb']['port'],
            args=(config.cmd_args.verbose,)
        ),
        TrunkProcessEnum.POSTGRES: TrunkProcessData(
            name=TrunkProcessEnum.POSTGRES.value,
            entrypoint=start_postgres,
            port=config['api']['postgres']['port'],
            args=(config.cmd_args.verbose,)
        ),
        TrunkProcessEnum.JOBS: TrunkProcessData(
            name=TrunkProcessEnum.JOBS.value,
            entrypoint=start_scheduler,
            args=(config.cmd_args.verbose,)
        ),
        TrunkProcessEnum.TASKS: TrunkProcessData(
            name=TrunkProcessEnum.TASKS.value,
            entrypoint=start_tasks,
            args=(config.cmd_args.verbose,)
        ),
        TrunkProcessEnum.ML_TASK_QUEUE: TrunkProcessData(
            name=TrunkProcessEnum.ML_TASK_QUEUE.value,
            entrypoint=start_ml_task_queue,
            args=(config.cmd_args.verbose,)
        )
    }

    for api_enum in api_arr:
        trunc_processes_struct[api_enum].need_to_run = True

    if config['jobs']['disable'] is False:
        trunc_processes_struct[TrunkProcessEnum.JOBS].need_to_run = True

    if config['tasks']['disable'] is False:
        trunc_processes_struct[TrunkProcessEnum.TASKS].need_to_run = True

    if config.cmd_args.ml_task_queue_consumer is True:
        trunc_processes_struct[TrunkProcessEnum.ML_TASK_QUEUE].need_to_run = True

    def start_process(trunc_process_data):
        # TODO this 'ctx' is eclipsing 'context' class imported as 'ctx'
        ctx = mp.get_context("spawn")
        logger.info(f"{trunc_process_data.name} API: starting...")
        try:
            trunc_process_data.process = ctx.Process(
                target=trunc_process_data.entrypoint,
                args=trunc_process_data.args,
                name=trunc_process_data.name
            )
            trunc_process_data.process.start()
        except Exception as e:
            logger.error(
                f"Failed to start {trunc_process_data.name} API with exception {e}\n{traceback.format_exc()}"
            )
            close_api_gracefully(trunc_processes_struct)
            raise e

    for trunc_process_data in trunc_processes_struct.values():
        if trunc_process_data.started is True or trunc_process_data.need_to_run is False:
            continue
        start_process(trunc_process_data)

    atexit.register(close_api_gracefully, trunc_processes_struct=trunc_processes_struct)

    async def wait_api_start(api_name, pid, port):
        timeout = 60
        start_time = time.time()
        started = is_pid_listen_port(pid, port)
        while (time.time() - start_time) < timeout and started is False:
            await asyncio.sleep(0.5)
            started = is_pid_listen_port(pid, port)
        return api_name, port, started

    async def wait_apis_start():
        futures = [
            wait_api_start(
                trunc_process_data.name,
                trunc_process_data.process.pid,
                trunc_process_data.port
            )
            for trunc_process_data in trunc_processes_struct.values()
            if trunc_process_data.port is not None and trunc_process_data.need_to_run is True
        ]
        for future in asyncio.as_completed(futures):
            api_name, port, started = await future
            if started:
                logger.info(f"{api_name} API: started on {port}")
            else:
                logger.error(f"ERROR: {api_name} API cant start on {port}")

    async def join_process(trunc_process_data: TrunkProcessData):
        finish = False
        while not finish:
            process = trunc_process_data.process
            try:
                while process.is_alive():
                    process.join(1)
                    await asyncio.sleep(0)
            except KeyboardInterrupt:
                logger.info("Got keyboard interrupt, stopping APIs")
                close_api_gracefully(trunc_processes_struct)
            finally:
                if trunc_process_data.should_restart:
                    if trunc_process_data.request_restart_attempt():
                        logger.warning(f"{trunc_process_data.name} API: stopped unexpectedly, restarting")
                        trunc_process_data.process = None
                        if trunc_process_data.name == TrunkProcessEnum.HTTP.value:
                            # do not open GUI on HTTP API restart
                            trunc_process_data.args = (config.cmd_args.verbose, True)
                        start_process(trunc_process_data)
                        api_name, port, started = await wait_api_start(
                            trunc_process_data.name,
                            trunc_process_data.process.pid,
                            trunc_process_data.port
                        )
                        if started:
                            logger.info(f"{api_name} API: started on {port}")
                        else:
                            logger.error(f"ERROR: {api_name} API cant start on {port}")
                    else:
                        finish = True
                        logger.error(
                            f'The "{trunc_process_data.name}" process could not restart after failure. '
                            'There will be no further attempts to restart.'
                        )
                else:
                    finish = True
                    logger.info(f"{trunc_process_data.name} API: stopped")

    async def gather_apis():
        await asyncio.gather(
            *[
                join_process(trunc_process_data)
                for trunc_process_data in trunc_processes_struct.values()
                if trunc_process_data.need_to_run is True
            ],
            return_exceptions=False
        )

    ioloop = asyncio.new_event_loop()
    ioloop.run_until_complete(wait_apis_start())

    threading.Thread(target=do_clean_process_marks).start()

    ioloop.run_until_complete(gather_apis())
    ioloop.close()



from bson.int64 import Int64

from mindsdb_sql_parser.ast import Join, Select, Identifier, Describe, Show, Constant
import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes import Responder
from mindsdb.api.mongo.utilities.mongodb_ast import MongoToAst
from mindsdb.interfaces.jobs.jobs_controller import JobsController

from mindsdb.api.mongo.classes.query_sql import run_sql_command


def find_to_ast(query, database):
    mongoToAst = MongoToAst()

    collection = [database, query['find']]

    if not query.get('singleBatch') and 'collection' in query.get('filter'):
        # JOIN mode

        # upper query
        ast_query = mongoToAst.find(
            collection=collection,
            projection=query.get('projection'),
            sort=query.get('sort'),
            limit=query.get('limit'),
            skip=query.get('skip'),
        )

        # table_query
        collection = query['filter']['collection']
        filter = query['filter'].get('query', {})
        table_select = mongoToAst.find(
            collection=collection,
            filter=filter,
        )
        table_select.parentheses = True
        table_select.alias = Identifier(query['filter']['collection'])
        table_select.alias.parts = [table_select.alias.parts[-1]]
        if 'limit' in query:
            table_select.limit = Constant(query['limit'])

        modifiers = query['filter'].get('modifiers')
        if modifiers is not None and hasattr(ast_query, 'modifiers'):
            for modifier in modifiers:
                table_select.modifiers.append(modifier)

        # convert to join
        right_table = ast_query.from_table

        ast_join = Join(
            left=table_select,
            right=right_table,
            join_type='join'
        )
        ast_query.from_table = ast_join

    else:
        # is single table
        ast_query = mongoToAst.find(
            collection=collection,
            filter=query.get('filter'),
            projection=query.get('projection'),
            sort=query.get('sort'),
            limit=query.get('limit'),
            skip=query.get('skip'),
        )
        modifiers = query['filter'].get('modifiers')
        if modifiers is not None and hasattr(ast_query, 'modifiers'):
            for modifier in modifiers:
                ast_query.modifiers.append(modifier)
    return ast_query


class Responce(Responder):
    when = {'find': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        database = request_env['database']
        project_name = request_env['database']

        if database == 'config':
            # return nothing
            return {
                'cursor': {
                    'id': Int64(0),
                    'ns': f"{database}.$cmd.{query['find']}",
                    'firstBatch': []
                },
                'ok': 1
            }

        # system queries
        elif query['find'] == 'system.version':
            # For studio3t
            data = [{
                "_id": "featureCompatibilityVersion",
                "version": "3.6"
            }]
            cursor = {
                'id': Int64(0),
                'ns': f"system.version.$cmd.{query['find']}",
                'firstBatch': data
            }
            return {
                'cursor': cursor,
                'ok': 1
            }

        elif query['find'] == 'ml_engines':
            ast_query = Show('ml_engines')

        else:
            ast_query = find_to_ast(query, database)

        # add _id for objects
        table_name = None
        obj_idx = {}
        if isinstance(ast_query, Select) and ast_query.from_table is not None:
            if isinstance(ast_query.from_table, Identifier):
                table_name = ast_query.from_table.parts[-1].lower()

                if table_name == 'models':

                    models = mindsdb_env['model_controller'].get_models(
                        ml_handler_name=None,
                        project_name=project_name
                    )

                    for model in models:
                        obj_idx[model['name']] = model['id']

                    # is select from model without where
                    if table_name in obj_idx and ast_query.where is None:
                        # replace query to describe model
                        ast_query = Describe(ast_query.from_table)
                elif table_name == 'jobs':
                    jobs_controller = JobsController()
                    for job in jobs_controller.get_list(project_name):
                        obj_idx[job['name']] = job['id']

        data = run_sql_command(request_env, ast_query)

        if table_name == 'models':
            # for models and models_versions _id is:
            #   - first 20 bytes is version
            #   - next bytes is model id

            for row in data:
                model_id = obj_idx.get(row.get('NAME'))
                if model_id is not None:
                    obj_id = model_id << 20

                    obj_id += row.get('VERSION', 0)

                    row['_id'] = helpers.int_to_objectid(obj_id)
        elif table_name == 'jobs':
            for row in data:
                obj_id = obj_idx.get(row.get('NAME'))
                if obj_id is not None:
                    row['_id'] = helpers.int_to_objectid(obj_id)

        db = mindsdb_env['config']['api']['mongodb']['database']

        cursor = {
            'id': Int64(0),
            'ns': f"{db}.$cmd.{query['find']}",
            'firstBatch': data
        }
        return {
            'cursor': cursor,
            'ok': 1
        }


responder = Responce()



import os
import json
from enum import Enum
from typing import Dict, List, Union
from urllib.parse import urlparse

import pandas as pd
import psycopg
from mindsdb_sql_parser.ast import Parameter, Identifier, Update, BinaryOperation
from pgvector.psycopg import register_vector

from mindsdb.integrations.handlers.postgres_handler.postgres_handler import (
    PostgresHandler,
)
from mindsdb.integrations.libs.response import RESPONSE_TYPE, HandlerResponse as Response
from mindsdb.integrations.libs.vectordatabase_handler import (
    FilterCondition,
    VectorStoreHandler,
    DistanceFunction,
    TableField
)
from mindsdb.utilities import log
from mindsdb.utilities.profiler import profiler
from mindsdb.utilities.context import context as ctx

logger = log.getLogger(__name__)


# todo Issue #7316 add support for different indexes and search algorithms e.g. cosine similarity or L2 norm
class PgVectorHandler(VectorStoreHandler, PostgresHandler):
    """This handler handles connection and execution of the PostgreSQL with pgvector extension statements."""

    name = "pgvector"

    def __init__(self, name: str, **kwargs):

        super().__init__(name=name, **kwargs)
        self._is_shared_db = False
        self._is_vector_registered = False
        # we get these from the connection args on PostgresHandler parent
        self._is_sparse = self.connection_args.get('is_sparse', False)
        self._vector_size = self.connection_args.get('vector_size', None) 
        if self._is_sparse and not self._vector_size:
            raise ValueError("vector_size is required when is_sparse=True")
        self.connect()

    def _make_connection_args(self):
        cloud_pgvector_url = os.environ.get('KB_PGVECTOR_URL')
        if cloud_pgvector_url is not None:
            result = urlparse(cloud_pgvector_url)
            self.connection_args = {
                'host': result.hostname,
                'port': result.port,
                'user': result.username,
                'password': result.password,
                'database': result.path[1:]
            }
            self._is_shared_db = True
        return super()._make_connection_args()

    def get_tables(self) -> Response:
        # Hide list of tables from all users
        if self._is_shared_db:
            return Response(RESPONSE_TYPE.OK)
        return super().get_tables()

    def native_query(self, query) -> Response:
        # Prevent execute native queries
        if self._is_shared_db:
            return Response(RESPONSE_TYPE.OK)
        return super().native_query(query)

    def raw_query(self, query, params=None) -> Response:
        resp = super().native_query(query, params)
        if resp.resp_type == RESPONSE_TYPE.ERROR:
            raise RuntimeError(resp.error_message)
        if resp.resp_type == RESPONSE_TYPE.TABLE:
            return resp.data_frame

    @profiler.profile()
    def connect(self) -> psycopg.connection:
        """
        Handles the connection to a PostgreSQL database instance.
        """
        self.connection = super().connect()
        if self._is_vector_registered:
            return self.connection

        with self.connection.cursor() as cur:
            try:
                # load pg_vector extension
                cur.execute("CREATE EXTENSION IF NOT EXISTS vector")
                logger.info("pg_vector extension loaded")

            except psycopg.Error as e:
                self.connection.rollback()
                logger.error(
                    f"Error loading pg_vector extension, ensure you have installed it before running, {e}!"
                )
                raise

        # register vector type with psycopg2 connection
        register_vector(self.connection)
        self._is_vector_registered = True

        return self.connection

    @staticmethod
    def _translate_conditions(conditions: List[FilterCondition]) -> Union[dict, None]:
        """
        Translate filter conditions to a dictionary
        """

        if conditions is None:
            return {}

        return {
            condition.column.split(".")[-1]: {
                "op": condition.op.value,
                "value": condition.value,
            }
            for condition in conditions
        }

    @staticmethod
    def _construct_where_clause(filter_conditions=None):
        """
        Construct where clauses from filter conditions
        """
        if filter_conditions is None:
            return ""

        where_clauses = []

        for key, value in filter_conditions.items():
            if key == "embeddings":
                continue
            if value['op'].lower() == 'in':
                values = list(repr(i) for i in value['value'])
                value['value'] = '({})'.format(', '.join(values))
            else:
                value['value'] = repr(value['value'])
            where_clauses.append(f'{key} {value["op"]} {value["value"]}')

        if len(where_clauses) > 1:
            return f"WHERE{' AND '.join(where_clauses)}"
        elif len(where_clauses) == 1:
            return f"WHERE {where_clauses[0]}"
        else:
            return ""

    @staticmethod
    def _construct_full_after_from_clause(
        offset_clause: str,
        limit_clause: str,
        where_clause: str,
    ) -> str:

        return f"{where_clause} {offset_clause} {limit_clause}"

    def _build_select_query(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        limit: int = None,
        offset: int = None,
    ) -> str:
        """
        given inputs, build string query
        """
        limit_clause = f"LIMIT {limit}" if limit else ""
        offset_clause = f"OFFSET {offset}" if offset else ""

        # translate filter conditions to dictionary
        filter_conditions = self._translate_conditions(conditions)

        # check if search vector is in filter conditions
        embedding_search = filter_conditions.get("embeddings", None)

        # given filter conditions, construct where clause
        where_clause = self._construct_where_clause(filter_conditions)

        # construct full after from clause, where clause + offset clause + limit clause
        after_from_clause = self._construct_full_after_from_clause(
            where_clause, offset_clause, limit_clause
        )

        if columns is None:
            targets = '*'
        else:
            targets = ', '.join(columns)


        if filter_conditions:

            if embedding_search:
                search_vector = filter_conditions["embeddings"]["value"][0]
                filter_conditions.pop("embeddings")

                if self._is_sparse:
                    # Convert dict to sparse vector if needed
                    if isinstance(search_vector, dict):
                        from pgvector.utils import SparseVector
                        embedding = SparseVector(search_vector, self._vector_size)
                        search_vector = embedding.to_text()
                    # Use inner product for sparse vectors
                    distance_op = "<#>"
                else:
                    # Convert list to vector string if needed
                    if isinstance(search_vector, list):
                        search_vector = f"[{','.join(str(x) for x in search_vector)}]"
                    # Use cosine similarity for dense vectors
                    distance_op = "<=>"

                return f"SELECT {targets} FROM {table_name} ORDER BY embeddings {distance_op} '{search_vector}' ASC {after_from_clause}"

            else:
                # if filter conditions, return rows that satisfy the conditions
                return f"SELECT {targets} FROM {table_name} {after_from_clause}"

        else:
            # if no filter conditions, return all rows
            return f"SELECT {targets} FROM {table_name} {after_from_clause}"

    def _check_table(self, table_name: str):
        # Apply namespace for a user
        if self._is_shared_db:
            company_id = ctx.company_id or 'x'
            return f't_{company_id}_{table_name}'
        return table_name

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> pd.DataFrame:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        table_name = self._check_table(table_name)

        if columns is None:
            columns = ["id", "content", "embeddings", "metadata"]

        query = self._build_select_query(table_name, columns, conditions, limit, offset)
        result = self.raw_query(query)

        # ensure embeddings are returned as string so they can be parsed by mindsdb
        if "embeddings" in columns:
            result["embeddings"] = result["embeddings"].astype(str)

        return result

    def hybrid_search(
        self,
        table_name: str,
        embeddings: List[float],
        query: str = None,
        metadata: Dict[str, str] = None,
        distance_function = DistanceFunction.COSINE_DISTANCE,
        **kwargs
    ) -> pd.DataFrame:
        '''
        Executes a hybrid search, combining semantic search and one or both of keyword/metadata search.

        For insight on the query construction, see: https://docs.pgvecto.rs/use-case/hybrid-search.html#advanced-search-merge-the-results-of-full-text-search-and-vector-search.

        Args:
            table_name(str): Name of underlying table containing content, embeddings, & metadata
            embeddings(List[float]): Embedding vector to perform semantic search against
            query(str): User query to convert into keywords for keyword search
            metadata(Dict[str, str]): Metadata filters to filter content rows against
            distance_function(DistanceFunction): Distance function used to compare embeddings vectors for semantic search

        Kwargs:
            id_column_name(str): Name of ID column in underlying table
            content_column_name(str): Name of column containing document content in underlying table
            embeddings_column_name(str): Name of column containing embeddings vectors in underlying table
            metadata_column_name(str): Name of column containing metadata key-value pairs in underlying table

        Returns:
            df(pd.DataFrame): Hybrid search result, sorted by hybrid search rank
        '''
        if query is None and metadata is None:
            raise ValueError('Must provide at least one of: query for keyword search, or metadata filters. For only embeddings search, use normal search instead.')

        id_column_name = kwargs.get('id_column_name', 'id')
        content_column_name = kwargs.get('content_column_name', 'content')
        embeddings_column_name = kwargs.get('embeddings_column_name', 'embeddings')
        metadata_column_name = kwargs.get('metadata_column_name', 'metadata')
        # Filter by given metadata for semantic search & full text search CTEs, if present.
        where_clause = ' WHERE '
        if metadata is None:
            where_clause = ''
            metadata = {}
        for i, (k, v) in enumerate(metadata.items()):
            where_clause += f"{metadata_column_name}->>'{k}' = '{v}'"
            if i < len(metadata.items()) - 1:
                where_clause += ' AND '

        # See https://docs.pgvecto.rs/use-case/hybrid-search.html#advanced-search-merge-the-results-of-full-text-search-and-vector-search.
        #
        # We can break down the below query as follows:
        # 
        # Start with a CTE (Common Table Expression) called semantic_search (https://www.postgresql.org/docs/current/queries-with.html).
        # This expression calculates rank by the defined distance function, which measures the distance between the
        # embeddings column and the given embeddings vector. Results are ordered by this rank.
        #
        # Next, define another CTE called full_text_search if we are doing keyword search.
        # This calculates rank using the built-in ts_rank function (https://www.postgresql.org/docs/current/textsearch-controls.html#TEXTSEARCH-RANKING).
        # We convert the content column to a ts_vector and match rows for the given tsquery in the content column. Results are ordered by this ts_rank.
        #
        # For both of these CTEs, we filter by any given metadata fields.
        #
        # See https://www.postgresql.org/docs/current/textsearch-controls.html#TEXTSEARCH-PARSING-DOCUMENTS for to_tsvector
        # See https://www.postgresql.org/docs/current/functions-textsearch.html#FUNCTIONS-TEXTSEARCH for tsquery syntax
        #
        # Finally, we use a FULL OUTER JOIN to SELECT from both CTEs defined above.
        # The COALESCE function is used to handle cases where one CTE has null values.
        #
        # Or, if we are only doing metadata search, we leave out the JOIN & full text search CTEs.
        #
        # We calculate the final "hybrid" rank by summing the reciprocals of the ranks from each individual CTE.
        semantic_search_cte = f'''WITH semantic_search AS (
    SELECT {id_column_name}, {content_column_name}, {embeddings_column_name},
    RANK () OVER (ORDER BY {embeddings_column_name} {distance_function.value} '{str(embeddings)}') AS rank
    FROM {table_name}{where_clause}
    ORDER BY {embeddings_column_name} {distance_function.value} '{str(embeddings)}'::vector
    )'''

        full_text_search_cte = ''
        if query is not None:
            ts_vector_clause = f"WHERE to_tsvector('english', {content_column_name}) @@ plainto_tsquery('english', '{query}')"
            if metadata:
                ts_vector_clause = f"AND to_tsvector('english', {content_column_name}) @@ plainto_tsquery('english', '{query}')"
            full_text_search_cte = f''',
    full_text_search AS (
    SELECT {id_column_name}, {content_column_name}, {embeddings_column_name},
    RANK () OVER (ORDER BY ts_rank(to_tsvector('english', {content_column_name}), plainto_tsquery('english', '{query}')) DESC) AS rank
    FROM {table_name}{where_clause}
    {ts_vector_clause}
    ORDER BY ts_rank(to_tsvector('english', {content_column_name}), plainto_tsquery('english', '{query}')) DESC
    )'''

        hybrid_select = '''
    SELECT * FROM semantic_search'''
        if query is not None:
            hybrid_select = f'''
    SELECT
        COALESCE(semantic_search.{id_column_name}, full_text_search.{id_column_name}) AS id,
        COALESCE(semantic_search.{content_column_name}, full_text_search.{content_column_name}) AS content,
        COALESCE(semantic_search.{embeddings_column_name}, full_text_search.{embeddings_column_name}) AS embeddings,
        COALESCE(1.0 / (1 + semantic_search.rank), 0.0) + COALESCE(1.0 / (1 + full_text_search.rank), 0.0) AS rank
    FROM semantic_search FULL OUTER JOIN full_text_search USING ({id_column_name}) ORDER BY rank DESC;
        '''

        full_search_query = f'{semantic_search_cte}{full_text_search_cte}{hybrid_select}'
        return self.raw_query(full_search_query)

    def create_table(self, table_name: str):
        """Create a table with a vector column."""
        with self.connection.cursor() as cur:
            # For sparse vectors, use sparsevec type
            vector_column_type = 'sparsevec' if self._is_sparse else 'vector'
            
            # Vector size is required for sparse vectors, optional for dense
            if self._is_sparse and not self._vector_size:
                raise ValueError("vector_size is required for sparse vectors")
            
            # Add vector size specification only if provided
            size_spec = f"({self._vector_size})" if self._vector_size is not None else "()"
            if vector_column_type == 'vector':
                size_spec = ''

            cur.execute(f"""
                CREATE TABLE IF NOT EXISTS {table_name} (
                    id TEXT PRIMARY KEY,
                    embeddings {vector_column_type}{size_spec},
                    content TEXT,
                    metadata JSONB
                )
            """)
            self.connection.commit()

    def insert(
        self, table_name: str, data: pd.DataFrame
    ):
        """
        Insert data into the pgvector table database.
        """
        table_name = self._check_table(table_name)

        data_dict = data.to_dict(orient="list")

        if 'metadata' in data_dict:
            data_dict['metadata'] = [json.dumps(i) for i in data_dict['metadata']]
        transposed_data = list(zip(*data_dict.values()))

        columns = ", ".join(data.keys())
        values = ", ".join(["%s"] * len(data.keys()))

        insert_statement = f"INSERT INTO {table_name} ({columns}) VALUES ({values})"

        self.raw_query(insert_statement, params=transposed_data)

    def update(
        self, table_name: str, data: pd.DataFrame, key_columns: List[str] = None
    ):
        """
        Udate data into the pgvector table database.
        """
        table_name = self._check_table(table_name)

        where = None
        update_columns = {}

        for col in data.columns:
            value = Parameter('%s')

            if col in key_columns:
                cond = BinaryOperation(
                    op='=',
                    args=[Identifier(col), value]
                )
                if where is None:
                    where = cond
                else:
                    where = BinaryOperation(
                        op='AND',
                        args=[where, cond]
                    )
            else:
                update_columns[col] = value

        query = Update(
            table=Identifier(table_name),
            update_columns=update_columns,
            where=where
        )

        if TableField.METADATA.value in data.columns:
            def fnc(v):
                if isinstance(v, dict):
                    return json.dumps(v)
            data[TableField.METADATA.value] = data[TableField.METADATA.value].apply(fnc)

            data = data.astype({TableField.METADATA.value: str})

        transposed_data = []
        for _, record in data.iterrows():
            row = [
                record[col]
                for col in update_columns.keys()
            ]
            for key_column in key_columns:
                row.append(record[key_column])
            transposed_data.append(row)

        query_str = self.renderer.get_string(query)
        self.raw_query(query_str, transposed_data)

    def delete(
        self, table_name: str, conditions: List[FilterCondition] = None
    ):
        table_name = self._check_table(table_name)

        filter_conditions = self._translate_conditions(conditions)
        where_clause = self._construct_where_clause(filter_conditions)

        query = (
            f"DELETE FROM {table_name} {where_clause}"
        )
        self.raw_query(query)

    def drop_table(self, table_name: str, if_exists=True):
        """
        Run a drop table query on the pgvector database.
        """
        table_name = self._check_table(table_name)
        self.raw_query(f"DROP TABLE IF EXISTS {table_name}")



import pandas as pd
import dill
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.integrations.utilities.time_series_utils import (
    transform_to_nixtla_df,
    get_results_from_nixtla_df,
    infer_frequency,
    get_best_model_from_results_df,
    get_model_accuracy_dict,
    reconcile_forecasts,
    get_hierarchy_from_df
)
from sklearn.metrics import r2_score
from statsforecast import StatsForecast
from statsforecast.models import AutoARIMA, AutoCES, AutoETS, AutoTheta

# hierarchicalforecast is an optional dependency
try:
    from hierarchicalforecast.core import HierarchicalReconciliation
except ImportError:
    HierarchicalReconciliation = None

DEFAULT_MODEL_NAME = "AutoARIMA"
model_dict = {
    "AutoARIMA": AutoARIMA,
    "AutoCES": AutoCES,
    "AutoETS": AutoETS,
    "AutoTheta": AutoTheta,
}


def get_season_length(frequency):
    """Infers best season length from frequency parameter.

    We set a sensible default for seasonality based on the
    frequency parameter. For example: we assume monthly data
    has a season length of 12 (months in a year). If the inferred frequency
    isn't found, we default to 1 i.e. no seasonality.
    """
    season_dict = {  # https://pandas.pydata.org/docs/user_guide/timeseries.html#timeseries-offset-aliases
        "H": 24,
        "M": 12,
        "MS": 12,
        "Q": 4,
        "SM": 24,
        "BM": 12,
        "BMS": 12,
        "BQ": 4,
        "BH": 24,
    }
    new_freq = frequency.split("-")[0] if "-" in frequency else frequency  # shortens longer frequencies like Q-DEC
    return season_dict[new_freq] if new_freq in season_dict else 1


def get_insample_cv_results(model_args, df):
    """Gets insample cross validation results"""
    season_length = get_season_length(model_args["frequency"]) if not model_args.get("season_length") else model_args["season_length"]  # noqa
    if model_args["model_name"] == "auto":
        models = [model(season_length=season_length) for model in model_dict.values()]
    else:
        models = [model_dict[model_args["model_name"]](season_length=season_length)]

    sf = StatsForecast(models, model_args["frequency"])
    sf.cross_validation(model_args["horizon"], df, fitted=True)
    results_df = sf.cross_validation_fitted_values()
    return results_df.rename({"CES": "AutoCES"}, axis=1)  # Fixes a Nixtla bug


def choose_model(model_args, results_df):
    """Chooses which model to use in StatsForecast.

    If the user passes 'auto' for their model_name, this will choose the best
    model based on in-sample cross validation performance. This will then modify
    the model_args dictionnary, replacing 'auto' with the best-performing model.
    """
    if model_args["model_name"] == "auto":
        model_args["model_name"] = get_best_model_from_results_df(results_df)
    model_args["season_length"] = get_season_length(model_args["frequency"]) if not model_args.get("season_length") else model_args["season_length"]  # noqa
    model = model_dict[model_args["model_name"]]
    return model(season_length=model_args["season_length"])


class StatsForecastHandler(BaseMLEngine):
    """Integration with the Nixtla StatsForecast library for
    time series forecasting with classical methods.
    """

    name = "statsforecast"

    def create(self, target, df, args={}):
        """Create the StatsForecast Handler.

        Requires specifying the target column to predict and time series arguments for
        prediction horizon, time column (order by) and grouping column(s).

        Saves args, models params, and the formatted training df to disk. The training df
        is used later in the predict() method.
        """
        time_settings = args["timeseries_settings"]
        using_args = args["using"]
        assert time_settings["is_timeseries"], "Specify time series settings in your query"
        ###### store model args and time series settings in the model folder
        model_args = {}
        model_args.update(using_args)
        model_args["target"] = target
        model_args["horizon"] = time_settings["horizon"]
        model_args["order_by"] = time_settings["order_by"]
        if 'group_by' not in time_settings:
            # add group column
            group_col = '__group_by'
            time_settings["group_by"] = [group_col]

        model_args["group_by"] = time_settings["group_by"]
        model_args["frequency"] = (
            using_args["frequency"] if "frequency" in using_args else infer_frequency(df, time_settings["order_by"])
        )
        model_args["hierarchy"] = using_args["hierarchy"] if "hierarchy" in using_args else False
        if model_args["hierarchy"] and HierarchicalReconciliation is not None:
            training_df, hier_df, hier_dict = get_hierarchy_from_df(df, model_args)
            self.model_storage.file_set("hier_dict", dill.dumps(hier_dict))
            self.model_storage.file_set("hier_df", dill.dumps(hier_df))
        else:
            training_df = transform_to_nixtla_df(df, model_args)

        model_args["model_name"] = DEFAULT_MODEL_NAME if "model_name" not in using_args else using_args["model_name"]

        results_df = get_insample_cv_results(model_args, training_df)
        model_args["accuracies"] = get_model_accuracy_dict(results_df, r2_score)
        model = choose_model(model_args, results_df)
        sf = StatsForecast([model], freq=model_args["frequency"], df=training_df)
        fitted_models = sf.fit().fitted_

        ###### persist changes to handler folder
        self.model_storage.json_set("model_args", model_args)
        self.model_storage.file_set("training_df", dill.dumps(training_df))
        self.model_storage.file_set("fitted_models", dill.dumps(fitted_models))

    def predict(self, df, args={}):
        """Makes forecasts with the StatsForecast Handler.

        StatsForecast is setup to predict for all groups, so it won't handle
        a dataframe that's been filtered to one group very well. Instead, we make
        the prediction for all groups then take care of the filtering after the
        forecasting. Prediction is nearly instant.
        """
        # Load model arguments
        model_args = self.model_storage.json_get("model_args")
        training_df = dill.loads(self.model_storage.file_get("training_df"))
        fitted_models = dill.loads(self.model_storage.file_get("fitted_models"))

        prediction_df = transform_to_nixtla_df(df, model_args)
        groups_to_keep = prediction_df["unique_id"].unique()

        sf = StatsForecast(models=[], freq=model_args["frequency"], df=training_df)
        sf.fitted_ = fitted_models
        model_name = str(fitted_models[0][0])
        forecast_df = sf.predict(model_args["horizon"])
        forecast_df.index = forecast_df.index.astype(str)

        if model_args["hierarchy"] and HierarchicalReconciliation is not None:
            hier_df = dill.loads(self.model_storage.file_get("hier_df"))
            hier_dict = dill.loads(self.model_storage.file_get("hier_dict"))
            reconciled_df = reconcile_forecasts(training_df, forecast_df, hier_df, hier_dict)
            results_df = reconciled_df[reconciled_df.index.isin(groups_to_keep)]

        else:
            results_df = forecast_df[forecast_df.index.isin(groups_to_keep)]

        result = get_results_from_nixtla_df(results_df, model_args)
        result = result.rename(columns={model_name: model_args['target']})
        return result

    def describe(self, attribute=None):
        model_args = self.model_storage.json_get("model_args")

        if attribute == "model":
            return pd.DataFrame({k: [model_args[k]] for k in ["model_name", "frequency", "season_length", "hierarchy"]})

        elif attribute == "features":
            return pd.DataFrame(
                {"ds": [model_args["order_by"]], "y": model_args["target"], "unique_id": [model_args["group_by"]]}
            )

        elif attribute == 'info':
            outputs = model_args["target"]
            inputs = [model_args["target"], model_args["order_by"], model_args["group_by"]]
            accuracies = [(model, acc) for model, acc in model_args["accuracies"].items()]
            return pd.DataFrame({"accuracies": [accuracies], "outputs": outputs, "inputs": [inputs]})

        else:
            tables = ['info', 'features', 'model']
            return pd.DataFrame(tables, columns=['tables'])



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'endSessions': helpers.is_true}

    result = {
        'ok': 1
    }


responder = Responce()



from typing import Dict, Optional

import pandas as pd
from portkey_ai import Portkey

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log

from mindsdb.integrations.utilities.handler_utils import get_api_key

logger = log.getLogger(__name__)

DEFAULT_METADATA = {
    "_source": "portkey-mindsdb-integration",
}


class PortkeyHandler(BaseMLEngine):
    """
    Integration with the Portkey LLM Python Library
    """

    name = "portkey"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True

    def create(
        self,
        target: str,
        df: Optional[pd.DataFrame] = None,
        args: Optional[Dict] = None,
    ) -> None:

        if "using" not in args:
            raise Exception(
                "Portkey engine requires a USING clause! Refer to its documentation for more details."
            )

        self.model_storage.json_set("args", args)

    def predict(
        self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None
    ) -> None:

        args = self.model_storage.json_get("args")
        api_key = get_api_key('portkey', args["using"], self.engine_storage, strict=False)

        self.client = Portkey(
            **self.engine_storage.get_connection_args(),
            api_key=api_key,
            metadata=DEFAULT_METADATA
        )

        result_df = pd.DataFrame()

        result_df["predictions"] = df["question"].apply(self._predict_answer)

        result_df = result_df.rename(columns={"predictions": args["target"]})

        return result_df

    def _predict_answer(self, text):
        """
        connects with portkey messages api to predict the answer for the particular question

        """

        model_args = self.model_storage.json_get("args")

        message = self.client.chat.completions.create(
            **model_args,
            messages=[
                {"role": "user", "content": text}
            ]
        )

        return message.choices[0].message.content
